"""
 This module replaces bodys with aggregates of the form X = #sum{}
 with an order encoding, if part of the predicate that is used inside the aggregate
 as an at most one restriction somewhere else in the program.
"""
import logging
from collections import defaultdict
from typing import Optional, cast

from clingo.ast import (
    AST,
    AggregateFunction,
    ASTType,
    BinaryOperation,
    BinaryOperator,
    Function,
    Literal,
    Sign,
    SymbolicAtom,
    UnaryOperation,
    UnaryOperator,
    Variable,
)
from clingo.symbol import SymbolType

from ngo.dependency import DomainPredicates, RuleDependency
from ngo.utils.ast import (
    LOC,
    AggAnalytics,
    AnnotatedPredicate,
    Predicate,
    collect_ast,
    collect_binding_information_body,
    is_conditional,
    is_predicate,
    loc2str,
    potentially_unifying_sequence,
    predicates,
)
from ngo.utils.globals import PREV, UniqueNames

log = logging.getLogger(__name__)


class SumAggregator:
    """Translates some predicates inside sum aggregates into chains"""

    def __init__(self, prg: list[AST], input_predicates: list[Predicate]):
        self.unique_names = UniqueNames(prg, input_predicates)
        self.input_predicates: set[Predicate] = set(input_predicates)
        self.rule_dependency = RuleDependency(prg)
        self.domain_predicates = DomainPredicates(self.unique_names, prg)
        # list of ({AggregateFunction.Max, AggregateFunction.Min}, Translation, index)
        #  where index is the position of the variable indicating the minimum/maximum
        self._atmost_preds: list[AnnotatedPredicate]
        self._atleast_preds: list[AnnotatedPredicate]
        self._atmost_preds, self._atleast_preds = self._calc_at_most(prg)
        self.objectives: dict[tuple[AST, ...], list[AST]] = defaultdict(list)
        self._collect_objectives(prg)

    def _collect_objectives(self, prg: list[AST]) -> None:
        for rule in prg:
            if rule.ast_type == ASTType.Minimize:
                self.objectives[
                    (
                        rule.weight,
                        rule.priority,
                        *rule.terms,
                    )
                ].append(rule)

    def _calc_at_most_on_rule(self, rule: AST) -> tuple[list[AnnotatedPredicate], list[AnnotatedPredicate]]:
        ret: tuple[list[AnnotatedPredicate], list[AnnotatedPredicate]] = ([], [])
        assert rule.ast_type == ASTType.Rule
        head: AST = rule.head
        body: list[AST] = rule.body
        if (
            not (
                head.ast_type == ASTType.HeadAggregate
                and head.function
                in (
                    AggregateFunction.Count,
                    AggregateFunction.Sum,
                )
            )
            and not head.ast_type == ASTType.Aggregate
        ):
            return ret

        analytics = AggAnalytics(head)
        if not analytics.guaranteed_leq(1):
            return ret

        preds: set[AnnotatedPredicate] = set()
        global_vars = collect_binding_information_body(body)[0]
        alone = True
        for elem in head.elements:
            condition: AST
            if head.ast_type == ASTType.HeadAggregate:
                condition = elem.condition
                if not (
                    elem.terms
                    and len(elem.terms) > 0
                    and elem.terms[0].ast_type == ASTType.SymbolicTerm
                    and elem.terms[0].symbol.type == SymbolType.Number
                    and elem.terms[0].symbol.number > 0
                ):
                    alone = False
                    continue
            else:
                condition = elem

            if condition.literal.sign != Sign.NoSign or condition.literal.atom.ast_type != ASTType.SymbolicAtom:
                alone = False
                continue
            assert condition.literal.atom.symbol.ast_type == ASTType.Function

            sa = condition.literal.atom.symbol
            p = Predicate(sa.name, len(sa.arguments))
            unprojected: list[int] = []
            for index, arg in enumerate(sa.arguments):
                local_vars = set(collect_ast(arg, "Variable"))
                if not global_vars or not local_vars.issubset(global_vars):
                    unprojected.append(index)
            preds.add(AnnotatedPredicate(p, tuple(unprojected)))

        if len(preds) != 1:
            return ret

        ret[0].append(next(iter(preds)))
        if analytics.guaranteed_geq(1) and alone:
            ret[1].append(next(iter(preds)))
        return ret

    def _calc_at_most(self, prg: list[AST]) -> tuple[list[AnnotatedPredicate], list[AnnotatedPredicate]]:
        """returning a list of predicates where there exists at most one,
        and a subset of where additionally there exists also at least one"""
        ret: tuple[list[AnnotatedPredicate], list[AnnotatedPredicate]] = ([], [])
        global_preds: set[Predicate] = set()
        for stm in prg:
            global_preds.update([spred.pred for spred in predicates(stm)])
        global_preds -= self.input_predicates
        for pred in global_preds:
            rules = self.rule_dependency.get_rules_that_derive(pred)
            if len(rules) != 1:
                continue
            at_most, at_least = self._calc_at_most_on_rule(rules[0])
            ret[0].extend(at_most)
            ret[1].extend(at_least)

        return ret

    def at_most_one_predicates(self) -> list[AnnotatedPredicate]:
        """returns a list of predicates inlcuding their unprojected positions
        where I'm sure that there exists at most one"""
        return self._atmost_preds

    def at_least_one_predicates(self) -> list[AnnotatedPredicate]:
        """returns a list of predicates inlcuding their unprojected positions
        where I'm sure that there exists at least one"""
        return self._atleast_preds

    def _get_trigger(self, minimize_var: AST, body: list[AST]) -> Optional[tuple[AST, int, AnnotatedPredicate]]:
        for lit in body:
            if is_conditional(lit):  # currently not supported, happens in soft constraints
                return None
            if not is_predicate(lit):
                continue
            symbol = lit.atom.symbol
            for next_anon_pred in self._atmost_preds:
                if next_anon_pred.pred == Predicate(symbol.name, len(symbol.arguments)):
                    anon_are_anonymous = True
                    for i in next_anon_pred.annotated_positions:
                        if symbol.arguments[i] == Variable(LOC, "_"):
                            continue
                        if symbol.arguments[i] == minimize_var:
                            trigger_index = i
                            continue
                        anon_are_anonymous = False
                    if anon_are_anonymous:
                        return (lit, trigger_index, next_anon_pred)
        return None

    @staticmethod
    def _element_passes(elem: AST, elements: list[AST]) -> bool:
        """True if element in sum aggregate is simple enough to be replaced inside chaining"""
        if elem.terms[0].ast_type != ASTType.Variable:  # only this variable as weight is allowed
            return False
        others = []
        for term in elem.terms[1:]:
            others.extend(collect_ast(term, "Variable"))
        for cond in elem.condition:
            others.extend(collect_ast(cond, "Variable"))
        if others.count(elem.terms[0]) != 1:
            return False
        if elem.condition is None:  # nocoverage # gringo optimizes this away
            return False
        for other in elements:
            if other == elem:
                continue
            if potentially_unifying_sequence(elem.terms, other.terms):
                return False
        return True

    def _replace_elements(self, elements: list[AST], prg: list[AST]) -> list[AST]:
        newelements = []
        for elem in elements:
            assert elem.ast_type == ASTType.BodyAggregateElement
            if elem.terms and len(elem.terms) > 0:
                if not self._element_passes(elem, elements):
                    newelements.append(elem)
                    continue

                trigger = self._get_trigger(elem.terms[0], elem.condition)

                if trigger is None:
                    newelements.append(elem)
                    continue
                trigger_lit, trigger_index, trigger_anon_pred = trigger
                log.info(f"Replace {trigger_anon_pred.pred.name}/{trigger_anon_pred.pred.arity} inside an aggregate.")

                old_condition = elem.condition
                old_condition.remove(trigger_lit)
                # a(X) :- X = #sum {L-P,D,L : chain(D,L), next(D,P,L); L,D,L: chain(D,L), not next(D,_,L) }.
                # 2. Compute domain/next predicate for origin of ProjectedPredicate
                prg.extend(self.domain_predicates.create_domain(trigger_anon_pred.pred))
                prg.extend(self.domain_predicates.create_next_pred_for_annotated_pred(trigger_anon_pred, trigger_index))
                prg.extend(
                    self.domain_predicates.create_chain_pred_for_annotated_pred(trigger_anon_pred, trigger_index, True)
                )
                chain_pred = self.domain_predicates.chain_pred(trigger_anon_pred, trigger_index, True)
                trigger_args = trigger_lit.atom.symbol.arguments
                var_global_flat = [
                    trigger_args[i]
                    for i in range(0, trigger_anon_pred.pred.arity)
                    if i not in trigger_anon_pred.annotated_positions
                ]
                var_l = trigger_args[trigger_index]
                new_symbol = trigger_lit.atom.symbol.update(name=chain_pred.name, arguments=var_global_flat + [var_l])
                new_atom = trigger_lit.atom.update(symbol=new_symbol)
                new_lit = trigger_lit.update(atom=new_atom)
                old_condition.append(new_lit)
                next_anotated_pred = self.domain_predicates.next_anon_predicate(trigger_anon_pred, trigger_index)
                new_condition = list(old_condition)
                new_condition.append(
                    Literal(
                        LOC,
                        Sign.NoSign,
                        SymbolicAtom(Function(LOC, next_anotated_pred.name, var_global_flat + [PREV, var_l], False)),
                    )
                )
                new_terms = list(elem.terms)
                new_terms[0] = BinaryOperation(LOC, BinaryOperator.Minus, elem.terms[0], PREV)
                var_global_flat_without_anon = [Variable(LOC, "none") if x.name == "_" else x for x in var_global_flat]
                new_terms.append(
                    Function(LOC, next_anotated_pred.name, var_global_flat_without_anon + [PREV, var_l], False)
                )
                newelements.append(elem.update(condition=new_condition, terms=new_terms))
                new_condition = list(old_condition)
                new_condition.append(
                    Literal(
                        LOC,
                        Sign.Negation,
                        SymbolicAtom(
                            Function(LOC, next_anotated_pred.name, var_global_flat + [Variable(LOC, "_"), var_l], False)
                        ),
                    )
                )
                new_terms[0] = elem.terms[0]
                new_terms.pop()
                new_terms.append(Function(LOC, next_anotated_pred.name, var_global_flat_without_anon + [var_l], False))
                newelements.append(elem.update(condition=new_condition, terms=new_terms))

        return newelements

    def _get_var(self, minimize: AST) -> Optional[AST]:
        """Returns the weight variable inside the minimization statement
        or none if the statement can not easily be replaced"""
        if minimize.body is None:
            return None  # nocoverage # gringo hides this from me
        unsafe: list[AST] = []
        for terms_, objective in self.objectives.items():
            if potentially_unifying_sequence(terms_, [minimize.weight, minimize.priority, *minimize.terms]):
                unsafe.extend([x for x in objective if x != minimize])
        if unsafe:
            log.info(f"Cannot optimize {loc2str(minimize.location)} as the tuple is not globally unique.")
            return None
        if minimize.weight.ast_type == ASTType.Variable:
            return cast(AST, minimize.weight)
        if (
            minimize.weight.ast_type == ASTType.UnaryOperation
            and minimize.weight.operator_type == UnaryOperator.Minus
            and minimize.weight.argument.ast_type == ASTType.Variable
        ):
            return cast(AST, minimize.weight.argument)
        return None

    def _replace_optimize(self, minimize: AST) -> list[AST]:
        ### very similar to bodyagg, but only one element.
        # Therefore we have body_literals instead of literals in the condition
        assert minimize.ast_type == ASTType.Minimize
        minimize_var = self._get_var(minimize)
        if minimize_var is None:
            return [minimize]

        others = []
        for term in [minimize.priority] + list(minimize.terms):
            others.extend(collect_ast(term, "Variable"))
        for cond in minimize.body:
            others.extend(collect_ast(cond, "Variable"))
        if others.count(minimize_var) != 1:
            return [minimize]
        trigger = self._get_trigger(minimize_var, minimize.body)

        if trigger is None:
            return [minimize]
        trigger_lit, trigger_index, trigger_anon_pred = trigger
        log.info(f"Replace {trigger_anon_pred.pred.name}/{trigger_anon_pred.pred.arity} inside an objective function.")

        old_condition = minimize.body
        old_condition.remove(trigger_lit)
        prg: list[AST] = []
        # a(X) :- X = #sum {L-P,D,L : chain(D,L), next(D,P,L); L,D,L: chain(D,L), not next(D,_,L) }.
        # 2. Compute domain/next predicate for origin of ProjectedPredicate
        prg.extend(self.domain_predicates.create_domain(trigger_anon_pred.pred))
        prg.extend(self.domain_predicates.create_next_pred_for_annotated_pred(trigger_anon_pred, trigger_index))
        prg.extend(self.domain_predicates.create_chain_pred_for_annotated_pred(trigger_anon_pred, trigger_index, True))
        chain_pred = self.domain_predicates.chain_pred(trigger_anon_pred, trigger_index, True)
        trigger_args = trigger_lit.atom.symbol.arguments
        var_global_flat = [
            trigger_args[i]
            for i in range(0, trigger_anon_pred.pred.arity)
            if i not in trigger_anon_pred.annotated_positions
        ]
        var_l = trigger_args[trigger_index]
        new_symbol = trigger_lit.atom.symbol.update(name=chain_pred.name, arguments=var_global_flat + [var_l])
        new_atom = trigger_lit.atom.update(symbol=new_symbol)
        new_lit = trigger_lit.update(atom=new_atom)
        old_condition.append(new_lit)
        next_anotated_pred = self.domain_predicates.next_anon_predicate(trigger_anon_pred, trigger_index)
        new_condition = list(old_condition)
        new_condition.append(
            Literal(
                LOC,
                Sign.NoSign,
                SymbolicAtom(Function(LOC, next_anotated_pred.name, var_global_flat + [PREV, var_l], False)),
            )
        )
        # new_terms = list(elem.terms)
        # new_terms[0] = BinaryOperation(LOC, BinaryOperator.Minus, elem.terms[0], PREV)
        weight = BinaryOperation(LOC, BinaryOperator.Minus, minimize_var, PREV)
        if minimize.weight.ast_type != ASTType.Variable:
            weight = UnaryOperation(LOC, UnaryOperator.Minus, weight)
        terms: list[AST] = list(minimize.terms)
        var_global_flat_without_anon = [Variable(LOC, "none") if x.name == "_" else x for x in var_global_flat]
        terms.append(Function(LOC, next_anotated_pred.name, var_global_flat_without_anon + [PREV, var_l], False))
        prg.append(minimize.update(weight=weight, terms=terms, body=new_condition))
        new_condition = list(old_condition)
        new_condition.append(
            Literal(
                LOC,
                Sign.Negation,
                SymbolicAtom(
                    Function(LOC, next_anotated_pred.name, var_global_flat + [Variable(LOC, "_"), var_l], False)
                ),
            )
        )
        terms.pop()
        terms.append(Function(LOC, next_anotated_pred.name, var_global_flat_without_anon + [var_l], False))
        prg.append(minimize.update(terms=terms, body=new_condition))
        return prg

    def execute(self, prg: list[AST]) -> list[AST]:
        """
        replace sum aggregates with sum aggregates containing chaining rules
        also does this for minimize statements
        """
        ret: list[AST] = []
        for stm in prg:
            if stm.ast_type in (ASTType.Rule, ASTType.Minimize):
                newbody = []
                for blit in stm.body:
                    if blit.ast_type == ASTType.Literal:
                        atom = blit.atom
                        if atom.ast_type == ASTType.BodyAggregate and atom.function in (
                            AggregateFunction.Sum,
                            AggregateFunction.SumPlus,
                        ):
                            newatom = atom.update(elements=self._replace_elements(atom.elements, ret))
                            newbody.append(blit.update(atom=newatom))
                        else:
                            newbody.append(blit)
                    else:
                        newbody.append(blit)
                stm = stm.update(body=newbody)
            if stm.ast_type == ASTType.Minimize:
                ret.extend(self._replace_optimize(stm))
            else:
                ret.append(stm)

        return ret



""" test removal of superseeded literals """
import pytest
from clingo.ast import AST, parse_string

from ngo.cleanup import CleanupTranslator
from ngo.normalize import postprocess, preprocess
from ngo.utils.ast import Predicate


@pytest.mark.parametrize(
    "lhs, input_predicates, rhs",
    (
        (
            """
a.
            """,
            [],
            """#program base.
a.""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a(X,Y) :- b(X,Y), dom(X), dom(Y).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a(X,Y) :- b(X,Y).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
:~ b(X,Y), dom(X), dom(Y). [X@1]
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
:~ b(X,Y). [X@1]""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a(X,Y) :- b(X,Y), dom(X), not dom(Y).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a(X,Y) :- b(X,Y); not dom(Y).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
b(X,Y) :- dom(X), dom(Y), X+Y < 23.
a(X,Y) :- b(X,Y), dom(X), dom(Y).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
b(X,Y) :- dom(X); dom(Y); (X+Y) < 23.
a(X,Y) :- b(X,Y).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42, not foo(X).
b(X,Y) :- dom(X), dom(Y), X+Y < 23, not foo(X).
a(X,Y) :- b(X,Y), dom(X), dom(Y), not foo(X).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42; not foo(X).
b(X,Y) :- dom(X); dom(Y); (X+Y) < 23; not foo(X).
a(X,Y) :- b(X,Y).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a(X,Y) :- b(X,Y), dom(X), dom(Y).
{ c(X,Y) } :- a(X,Y), dom(X), dom(Y).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a(X,Y) :- b(X,Y).
{ c(X,Y) } :- a(X,Y).""",
        ),
        (
            """
b(X,Y); c :- dom(X), dom(Y), X+Y > 42.
a(X,Y) :- b(X,Y), dom(X), dom(Y).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y); c :- dom(X); dom(Y); (X+Y) > 42.
a(X,Y) :- b(X,Y).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a(X,Y) :- b(X,Y), dom(X), dom(Y).
{ c(X,Y); d(X,Y) } :- a(X,Y), dom(X), dom(Y).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a(X,Y) :- b(X,Y).
{ c(X,Y); d(X,Y) } :- a(X,Y).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
{a(X,Y) : b(X,Y)} :- dom(X), dom(Y).
            """,
            [Predicate("dom", 1)],
            # can't remove dom(X) and Y here because they could have function for other elements in the head
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
{ a(X,Y): b(X,Y) } :- dom(X); dom(Y).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a(X,Y) :- b(X,Y), dom(X), dom(Y).
2 = #sum{1: c(X,Y); 2: d(X,Y) } :- a(X,Y), dom(X), dom(Y).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a(X,Y) :- b(X,Y).
2 = #sum { 1: c(X,Y); 2: d(X,Y) } :- a(X,Y).""",
        ),
        (
            """
#show a/3.
            """,
            [Predicate("dom", 1)],
            """#program base.
#show a/3.""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a :- e(X,Y): b(X,Y), dom(X), dom(Y).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a :- e(X,Y): b(X,Y).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a :- e(X,Y): b(X*X,Y+13), dom(X*X), dom(Y+13).
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a :- e(X,Y): b((X*X),(Y+13)).""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a :- 42 = #sum {X,Y : b(X,Y), dom(X), dom(Y)}.
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a :- 42 = #sum { X,Y: b(X,Y) }.""",
        ),
        (
            """
b(X,Y) :- dom(X), dom(Y), X+Y > 42.
a :- 42 {c : b(X,Y), dom(X), dom(Y)}.
            """,
            [Predicate("dom", 1)],
            """#program base.
b(X,Y) :- dom(X); dom(Y); (X+Y) > 42.
a :- 42 <= #sum { 1,0,c: c, b(X,Y) }.""",
        ),
        (  # technicly not wrong but weird,
            # if there is a seed to seq this is broken except it superseeds task/1 too, like
            # seq(T,0) :- task(T). It is still valid to remove task from the last rule, as its only source is task
            """
foo(T,S) :- seq(T,S).                                                  
seq(T,(S+1)) :- task(T); foo(T,S).
            """,
            [],
            """#program base.
foo(T,S) :- seq(T,S).
seq(T,(S+1)) :- foo(T,S).""",
        ),
        (  # same predicate
            """
b(X,Y) :- dom(X,Y), dom(_,Y).
            """,
            [Predicate("dom", 2)],
            """#program base.
b(X,Y) :- dom(X,Y).""",
        ),
    ),
)
def test_cleanup_translation(lhs: str, input_predicates: list[Predicate], rhs: str) -> None:
    """test removal of superseeded literals on whole programs"""
    ast: list[AST] = []
    parse_string(lhs, ast.append)
    ast = preprocess(ast)
    clt = CleanupTranslator(input_predicates)
    ast = clt.execute(ast)
    ast = postprocess(ast)
    output = "\n".join(map(str, ast))
    assert rhs == output


@pytest.mark.parametrize(
    "lhs, input_predicates, rhs",
    (
        (
            """
{ a; b }.
a :- #true.
            """,
            [],
            """#program base.
{ a; b }.
a.""",
        ),
        (
            """
{ a; b }.
a :- #true, b.
            """,
            [],
            """#program base.
{ a; b }.
a :- b.""",
        ),
        (
            """
{ a; b }.
a :- not #false, b.
            """,
            [],
            """#program base.
{ a; b }.
a :- b.""",
        ),
        (
            """
{ a; b }.
a :- #false, b.
            """,
            [],
            """#program base.
{ a; b }.""",
        ),
        (
            """
{ a; b }.
a :- not #true, b.
            """,
            [],
            """#program base.
{ a; b }.""",
        ),
        (
            """
{ a; b }.
a(X) :- X = #sum {1: b; 2: #true; 3: #false}.
            """,
            [],
            """#program base.
{ a; b }.
a(X) :- X = #sum { 1: b; 2 }.""",
        ),
        (
            """
{ a; b }.
a :- b : #true.
            """,
            [],
            """#program base.
{ a; b }.
a :- b.""",
        ),
        (
            """
{ a; b }.
a :- b : #false.
            """,
            [],
            """#program base.
{ a; b }.
a.""",
        ),
        (
            """
{ a; b }.
#minimize {5: #true, a; 3: #false,b}.
            """,
            [],
            """#program base.
{ a; b }.
:~ a. [5@0]""",
        ),
    ),
)
def test_cleanup_booleans(lhs: str, input_predicates: list[Predicate], rhs: str) -> None:
    """test removal of superseeded literals on whole programs"""
    ast: list[AST] = []
    parse_string(lhs, ast.append)
    ast = preprocess(ast)
    clt = CleanupTranslator(input_predicates)
    ast = clt.execute(ast)
    ast = postprocess(ast)
    output = "\n".join(map(str, ast))
    assert rhs == output



""" test removal of superseeded literals """
from typing import Optional

import pytest
from clingo.ast import AST, parse_string

from ngo.normalize import postprocess, preprocess
from ngo.unused import UnusedTranslator
from ngo.utils.ast import Predicate


@pytest.mark.parametrize(
    "lhs, input_predicates, output_predicates, rhs",
    (
        (
            """
a.
            """,
            [],
            [],
            """#program base.""",
        ),
        (
            """
a.
            """,
            [],
            [Predicate("a", 0)],
            """#program base.
a.""",
        ),
        (
            """
{ a } :- b(X).
            """,
            [Predicate("b", 1)],
            [],
            """#program base.
{ a } :- b(_).""",
        ),
        (
            """
b(X) :- c(X).
{ a } :- b(X).
            """,
            [],
            [],
            """#program base.
b :- c(X).
{ a } :- b.""",
        ),
        (
            """
b(X) :- c(X).
#show b/1.
            """,
            [],
            [Predicate("b", 1)],
            """#program base.
b(X) :- c(X).
#show b/1.""",
        ),
        (
            """
b(X) :- c(X).
{ a } :- b(X).
            """,
            [Predicate("b", 1)],
            [],
            """#program base.
b(X) :- c(X).
{ a } :- b(_).""",
        ),
        (
            """
b(X) :- c(X).
{ a } :- b(X).
            """,
            [],
            [Predicate("b", 1)],
            """#program base.
b(X) :- c(X).
{ a } :- b(_).""",
        ),
        (
            """
b(X) :- c(X), x.
{ a } :- b(X).
            """,
            [Predicate("b", 0), Predicate("x", 0)],
            [],
            """#program base.
b1 :- c(X); x.
{ a } :- b1.""",
        ),
        (
            """
b(X,X,X+1,42) :- c(X), x.
b(X,X+1,23) :- c(X), x.
{ a } :- b(X,Y,Z,W).
{ a } :- b(X,Y,Z).
            """,
            [Predicate("b", 0), Predicate("x", 0)],
            [],
            """#program base.
b1 :- c(X); x; AUX = (X+1).
b2 :- c(X); x; AUX = (X+1).
{ a } :- b1.
{ a } :- b2.""",
        ),
        (
            """
#false :- not 1 <= { order(T,S) } <= 1; S = (1..N); task_nr(N).
            """,
            [],
            [],
            """#program base.
#false :- not 1 <= #sum { 1,0,order(T,S): order(T,S) } <= 1; S = (1..N); task_nr(N).""",
        ),
        (
            """
b(1) :- c(X).
{ a } :- b(1).
            """,
            [Predicate("c", 1)],
            [],
            """#program base.
b(1) :- c(_).
{ a } :- b(1).""",
        ),
    ),
)
def test_unused_translation(
    lhs: str, input_predicates: list[Predicate], output_predicates: list[Predicate], rhs: str
) -> None:
    """test removal of superseeded literals on whole programs"""
    ast: list[AST] = []
    parse_string(lhs, ast.append)
    ast = preprocess(ast)
    utr = UnusedTranslator(ast, input_predicates, output_predicates)
    ast = utr.execute(ast)
    ast = postprocess(ast)
    output = "\n".join(map(str, ast))
    assert rhs == output


@pytest.mark.parametrize(
    "lhs, input_predicates, output_predicates, rhs",
    (
        (
            """
a.
            """,
            [],
            [],
            """#program base.""",
        ),
        (
            """
a.
            """,
            [],
            [Predicate("a", 0)],
            """#program base.
a.""",
        ),
        (
            """
{ a } :- b(X).
            """,
            [Predicate("b", 1)],
            [],
            """#program base.
{ a } :- b(_).""",
        ),
        (
            """
b(X) :- c(X).
{ a } :- b(X).
            """,
            [],
            [],
            """#program base.
{ a } :- c.""",
        ),
        (
            """
b(X) :- c(X).
#show b/1.
            """,
            [],
            [Predicate("b", 1)],
            """#program base.
b(X) :- c(X).
#show b/1.""",
        ),
        (
            """
b(X) :- c(X).
{ a } :- b(X).
            """,
            [Predicate("b", 1)],
            [],
            """#program base.
b(X) :- c(X).
{ a } :- b(_).""",
        ),
        (
            """
b(X) :- c(X).
{ a } :- b(X).
            """,
            [],
            [Predicate("b", 1)],
            """#program base.
b(X) :- c(X).
{ a } :- b(_).""",
        ),
        (
            """
b(X) :- c(X), x.
{ a } :- b(X).
            """,
            [Predicate("b", 0), Predicate("x", 0)],
            [],
            """#program base.
b1 :- c; x.
{ a } :- b1.""",
        ),
        (
            """
b(X,X,X+1,42) :- c(X), x.
b(X,X+1,23) :- c(X), x.
{ a } :- b(X,Y,Z,W).
{ a } :- b(X,Y,Z).
            """,
            [Predicate("b", 0), Predicate("x", 0)],
            [],
            """#program base.
b1 :- c(X); x; _ = (X+1).
b2 :- c(X); x; _ = (X+1).
{ a } :- b1.
{ a } :- b2.""",
        ),
        (
            """
b(X,X,X+1,42) :- c(X).
b(X,X+1,23) :- c(X).
a :- b(X,Y,Z,W).
a :- b(X,Y,Z).
            """,
            [Predicate("b", 0)],
            [],
            """#program base.""",
        ),
        (
            """
b(X) :- c(X).
{ a } :- b(X).
foo :- a.
            """,
            [Predicate("c", 1)],
            [],
            """#program base.
b :- c(_).
{ a } :- b.""",
        ),
        (
            """
__dom_match(B) :- person(B).
__aux_1(W) :- __dom_match(W); 3 <= #count { M1: match(M1,W) }.
            """,
            [Predicate("person", 1)],
            [Predicate("__aux_1", 1)],
            """#program base.
__aux_1(W) :- person(W); 3 <= #sum+ { 1,M1: match(M1,W) }.""",
        ),
        (
            """
dur(J,M,D) :- duration(J,M,D).
job(J) :- dur(J,_,_).
machine(M) :- dur(_,M,_).
jobs(X) :- X = { job(J) }.
machines(M) :- machine(M); not machine((M+1)).
time(T) :- T = (1..X); X = #sum { D,J,M: dur(J,M,D)}.
{ slot(J,M,T): dur(J,M,_) } :- machine(M); time(T).
            """,
            [Predicate("person", 1)],
            [Predicate("__aux_1", 1)],
            """#program base.
machine(M) :- duration(_,M,_).
time(T) :- T = (1..X); X = #sum { D,J,M: duration(J,M,D) }.
{ slot(J,M,T): duration(J,M,_) } :- machine(M); time(T).""",
        ),
        (  # dont optimize complex relations
            """
b(X) :- a(X,X*2).
c(X) :- b(X).
            """,
            [Predicate("a", 1)],
            [Predicate("c", 1)],
            """#program base.
b(X) :- a(X,(X*2)).
c(X) :- b(X).""",
        ),
        (  # dont optimize complex relations in head
            """
b(X*X) :- a(X).
c(X) :- b(X).
            """,
            [Predicate("a", 1)],
            [Predicate("c", 1)],
            """#program base.
b((X*X)) :- a(X).
c(X) :- b(X).""",
        ),
        (  # anonymous variables
            """
b(X,X,A) :- a(X,_,f(A)).
c(X*Z) :- b(X,X,Z).
            """,
            [Predicate("a", 3)],
            [Predicate("c", 1)],
            """#program base.
c((X*Z)) :- a(X,_,f(Z)).""",
        ),
        (
            """
executionTime(J,M,T) :- duration(J,M,T).
task(T) :- executionTime(T,_,_).
machine(M) :- executionTime(_,M,_).
maxtime(T,T1) :- task(T); executionTime(T,R1,T1); 0 <= { executionTime(T,R2,T2): R2 != R1, T2 > T1 } <= 0.
time((0..(Max-1))) :- Max = #sum { T1,T,R1: executionTime(T,R1,T1) }.
task_nr(S) :- S = #sum { 1,task(T): task(T) }.
1 <= { order(T,S): S = (1..Nr), task_nr(Nr) } <= 1 :- task(T).
         """,
            [Predicate("duration", 3)],
            [],
            """#program base.
task(T) :- duration(T,_,_).
task_nr(S) :- S = #sum { 1,task(T): task(T) }.
1 <= { order(T,S): S = (1..Nr), task_nr(Nr) } <= 1 :- task(T).""",
        ),
        (
            """
n(-1;0;1).
            """,
            [],
            [],
            """#program base.""",
        ),
        (
            """
wall(X1,Y1,X,Y) :- wall(X,Y,X1,Y1).
#false :- at(X,Y,S); at(X1,Y1,(S+1)); S = (0..(MAX-1)); steps(MAX); wall(X,Y,X1,Y1).
            """,
            [],
            [],
            """#program base.
wall(X1,Y1,X,Y) :- wall(X,Y,X1,Y1).
#false :- at(X,Y,S); at(X1,Y1,(S+1)); S = (0..(MAX-1)); steps(MAX); wall(X,Y,X1,Y1).""",
        ),
    ),
)
def test_unused_translation_fixpoint(
    lhs: str, input_predicates: list[Predicate], output_predicates: list[Predicate], rhs: str
) -> None:
    """test removal of superseeded literals on whole programs"""
    new_ast: Optional[list[AST]] = []
    ast: list[AST] = []
    parse_string(lhs, ast.append)
    ast = preprocess(ast)
    while True:
        utr = UnusedTranslator(ast, input_predicates, output_predicates)
        new_ast = utr.execute(ast)
        if new_ast == ast:
            break
        ast = new_ast
    ast = postprocess(ast)
    output = "\n".join(map(str, ast))
    assert rhs == output



""" convert an AST to a simplified normal form """

from functools import partial
from typing import Iterable, Optional

from clingo.ast import (
    AST,
    AggregateFunction,
    ASTType,
    BodyAggregate,
    BodyAggregateElement,
    Comparison,
    ComparisonOperator,
    Function,
    Guard,
    Literal,
    Sign,
    SymbolicTerm,
    Variable,
)
from clingo.symbol import Number

from ngo.utils.ast import (
    LOC,
    collect_ast,
    comparison2comparisonlist,
    global_vars_inside_body,
    is_body_aggregate,
    is_comparison,
    is_conditional,
    is_predicate,
    replace_var_name,
    transform_ast,
)
from ngo.utils.globals import AUX_VAR, UniqueVariables


def expand_comparisons(stm: AST) -> AST:
    """given a top level statement in a logic program, return a stm with expanded comparisons
    WE CAN NOT DO THIS REPLACEMENT IN THE FRONT OF COMPARISONS
    OR IN THE HEAD OF A RULE, wait for general clingo preprocessor
    """
    if stm.ast_type in (ASTType.Rule, ASTType.Minimize):
        return stm.update(body=normalize_operators(stm.body))
    return stm


def _normalize_operators_condition(condition: list[AST]) -> list[AST]:
    new_condition: list[AST] = []
    for c in condition:
        if c.ast_type == ASTType.Literal and c.atom.ast_type == ASTType.Comparison:
            new_condition.extend(
                [
                    Literal(LOC, c.sign, Comparison(lhs, [Guard(cop, rhs)]))
                    for lhs, cop, rhs in comparison2comparisonlist(c.atom)
                ]
            )
            continue
        new_condition.append(c)
    return new_condition


def normalize_operators(literals: Iterable[AST]) -> list[AST]:
    """replace a list of literals with a new list where all comparisons are binary and not chained"""
    new_literals: list[AST] = []
    for lit in literals:
        new_lit = lit.update()
        if lit.ast_type == ASTType.ConditionalLiteral:
            new_literals.append(lit.update(condition=_normalize_operators_condition(lit.condition)))
            continue
        if lit.ast_type != ASTType.Literal:
            new_literals.append(new_lit)  # nocoverage
            continue  # nocoverage
        atom = lit.atom
        if atom.ast_type == ASTType.Comparison:
            new_literals.extend(
                [
                    Literal(LOC, lit.sign, Comparison(lhs, [Guard(cop, rhs)]))
                    for lhs, cop, rhs in comparison2comparisonlist(atom)
                ]
            )
            continue
        if atom.ast_type == ASTType.BodyAggregate:
            new_elements: list[AST] = []
            for elem in atom.elements:
                new_elements.append(elem.update(condition=_normalize_operators_condition(elem.condition)))
            new_atom = atom.update(elements=new_elements)
            new_lit = lit.update(atom=new_atom)
        new_literals.append(new_lit)
    return new_literals


def _convert_count_to_sum(agg: AST) -> AST:
    """convert body count aggregate to sum with weight 1"""
    assert agg.ast_type == ASTType.BodyAggregate
    new_elements: list[AST] = []
    for old_elem in agg.elements:
        terms: list[AST] = [SymbolicTerm(LOC, Number(1)), *old_elem.terms]
        new_elements.append(old_elem.update(terms=terms))
    return agg.update(function=AggregateFunction.SumPlus, elements=new_elements)


def _replace_anon(symbol: AST) -> AST:
    return transform_ast(
        symbol, "Variable", partial(replace_var_name, Variable(LOC, "_"), Function(LOC, "anon__ngo", [], False))
    )


def _exline_interval(elem: AST, unique_vars: UniqueVariables) -> AST:
    """get a conditional literal, remove intervals from the literal part by shifting it to a variable"""
    assigns: list[AST] = []

    def replace_interval(interval: AST) -> AST:
        """replace interval with fresh variable and store it away"""
        aux = unique_vars.make_unique(AUX_VAR)
        assigns.append(Literal(LOC, Sign.NoSign, Comparison(aux, [Guard(ComparisonOperator.Equal, interval)])))
        return aux

    elem = transform_ast(elem, "Interval", replace_interval)
    return elem.update(condition=list(elem.condition) + assigns)


def _convert_old_agg(agg: AST, unqiue_vars: UniqueVariables) -> AST:
    """transforms old style body aggregate to new sum aggregate with weights 1"""
    assert agg.ast_type == ASTType.Aggregate
    nm = {Sign.NoSign: 0, Sign.Negation: 1, Sign.DoubleNegation: 2}
    bm = {True: 0, False: 1}
    new_elements: list[AST] = []
    comparison_counter = 2

    def replace_with_new(var: AST) -> AST:
        if var.name == "_":
            return unqiue_vars.make_unique(AUX_VAR)
        return var

    for old_elem in agg.elements:
        terms: list[AST] = []
        atom = old_elem.literal.atom
        old_elem = _exline_interval(old_elem, unqiue_vars)
        new_literal = old_elem.literal
        terms.append(SymbolicTerm(LOC, Number(1)))
        terms.append(SymbolicTerm(LOC, Number(nm[old_elem.literal.sign])))
        if atom.ast_type == ASTType.Comparison:
            terms.append(SymbolicTerm(LOC, Number(comparison_counter)))
            comparison_counter += 1
            terms.extend(sorted(collect_ast(atom, "Variable")))
        elif atom.ast_type == ASTType.BooleanConstant:
            terms.append(SymbolicTerm(LOC, Number(bm[atom.value])))
            comparison_counter += 1
        elif atom.ast_type == ASTType.SymbolicAtom:
            if new_literal.sign in (Sign.NoSign, Sign.DoubleNegation):
                new_literal = transform_ast(new_literal, "Variable", replace_with_new)
            terms.append(_replace_anon(new_literal.atom.symbol))
        else:
            assert False, f"Invalid atom {atom}"
        new_elements.append(BodyAggregateElement(terms, [new_literal, *old_elem.condition]))
    return BodyAggregate(agg.location, agg.left_guard, AggregateFunction.Sum, new_elements, agg.right_guard)


def replace_old_aggregates(prg: Iterable[AST]) -> list[AST]:
    """replace all oldstyle Aggregate`s in the body by BodyAggregate Sum
    Also replace body count aggregates with sum aggregates with weight of 1"""
    newprg: list[AST] = []
    for stm in prg:
        if stm.ast_type not in (ASTType.Rule, ASTType.Minimize):
            newprg.append(stm)
            continue
        if not stm.body:
            newprg.append(stm)
            continue
        newbody: list[AST] = []
        unique_vars = UniqueVariables(stm)
        for blit in stm.body:
            if blit.ast_type == ASTType.Literal and blit.atom.ast_type == ASTType.Aggregate:
                newbody.append(blit.update(atom=_convert_old_agg(blit.atom, unique_vars)))
            elif (
                blit.ast_type == ASTType.Literal
                and blit.atom.ast_type == ASTType.BodyAggregate
                and blit.atom.function == AggregateFunction.Count
            ):
                newbody.append(blit.update(atom=_convert_count_to_sum(blit.atom)))

            else:
                newbody.append(blit)
        newprg.append(stm.update(body=newbody))
    return newprg


def exline_term(term: AST, unique_vars: UniqueVariables) -> tuple[AST, list[AST]]:
    """try to replace arithmetic functions inside Functionsymbols with AUX Variables
    return the changed term and the new assignments"""
    uv: AST
    if term.ast_type in (ASTType.BinaryOperation, ASTType.UnaryOperation):
        uv = unique_vars.make_unique(AUX_VAR)
        return uv, [Literal(LOC, Sign.NoSign, Comparison(uv, [Guard(ComparisonOperator.Equal, term)]))]
    return term, []


def exline_literal(lit: AST, unique_vars: UniqueVariables) -> tuple[AST, list[AST]]:
    """try to replace arithmetic functions inside Functionsymbols with AUX Variables
    return the changed literal and the new assignments"""
    if not is_predicate(lit) or collect_ast(lit, "Pool"):
        return lit, []
    ret: list[AST] = []
    symbol = lit.atom.symbol
    new_terms: list[AST] = []
    for term in symbol.arguments:
        new_term, conditions = exline_term(term, unique_vars)
        new_terms.append(new_term)
        ret.extend(conditions)
    new_lit = lit.update(atom=lit.atom.update(symbol=symbol.update(arguments=new_terms)))
    return new_lit, ret


def exline_minimize_terms(stm: AST) -> AST:
    """move complex arithmetic from the minimize objective and terms to the body"""
    uv = UniqueVariables(stm)
    new_term, conditions = exline_term(stm.weight, uv)
    stm = stm.update(weight=new_term, body=list(stm.body) + conditions)
    new_term, conditions = exline_term(stm.priority, uv)
    stm = stm.update(priority=new_term, body=list(stm.body) + conditions)
    new_terms = []
    new_conditions = []
    for t in stm.terms:
        new_term, conditions = exline_term(t, uv)
        new_terms.append(new_term)
        new_conditions.extend(conditions)

    return stm.update(terms=new_terms, body=list(stm.body) + new_conditions)


def exline_arithmetic_rule(stm: AST) -> AST:
    """try to replace arithmetic functions inside Functionsymbols with AUX Variables"""
    unique_vars = UniqueVariables(stm)
    if stm.ast_type == ASTType.Rule:
        new_head, body = exline_literal(stm.head, unique_vars)
        stm = stm.update(head=new_head, body=list(stm.body) + body)
    if stm.ast_type == ASTType.Minimize:
        stm = exline_minimize_terms(stm)
    if stm.ast_type in (ASTType.Rule, ASTType.Minimize):
        new_body: list[AST] = []
        for blit in stm.body:
            if blit.ast_type == ASTType.Literal:
                new_lit, body = exline_literal(blit, unique_vars)
                new_body.extend([new_lit] + body)
            elif blit.ast_type == ASTType.ConditionalLiteral:
                new_condition: list[AST] = []
                for c in blit.condition:
                    new_lit, body = exline_literal(c, unique_vars)
                    new_condition.extend([new_lit] + body)
                new_body.append(blit.update(condition=new_condition))
            else:
                new_body.append(blit)  # nocoverage
        stm = stm.update(body=new_body)
    return stm


def exline_arithmetic(prg: list[AST]) -> list[AST]:
    """try to replace arithmetic functions inside Functionsymbols with AUX Variables"""
    return [exline_arithmetic_rule(stm) for stm in prg]


def normalize(prg: Iterable[AST]) -> list[AST]:
    """
    normalize a logic program by
     - remove arithmetics from head literals into body assignments
     - replace old style aggregates with new ones
     - replace nary operators with binary operators
     - unpooling
    """
    new_prg: list[AST] = []
    prg = replace_old_aggregates(prg)
    for stm in prg:
        new_prg.append(expand_comparisons(stm))
    prg, new_prg = new_prg, []
    for stm in prg:
        new_prg.extend(stm.unpool())

    return new_prg


def _equality(lit: AST) -> Optional[tuple[AST, AST]]:
    """given a lit, if it is of a form similar to X = Y+3,
    return X and Y+3"""
    if is_comparison(lit) and not collect_ast(lit, "Pool") and not collect_ast(lit, "Interval"):
        atom = lit.atom
        if atom.term.ast_type == ASTType.Variable and len(atom.guards) == 1:
            if (atom.guards[0].comparison == ComparisonOperator.Equal and lit.sign == Sign.NoSign) or (
                atom.guards[0].comparison == ComparisonOperator.NotEqual and lit.sign == Sign.Negation
            ):
                var = atom.term
                rest = atom.guards[0].term
                if var.name == "_":
                    return None
                return var, rest
        elif len(atom.guards) == 1 and atom.guards[0].term.ast_type == ASTType.Variable:
            if (atom.guards[0].comparison == ComparisonOperator.Equal and lit.sign == Sign.NoSign) or (
                atom.guards[0].comparison == ComparisonOperator.NotEqual and lit.sign == Sign.Negation
            ):
                var = atom.guards[0].term
                rest = atom.term
                if var.name == "_":
                    return None
                return var, rest
    return None


def inline_replace_stm(lit: AST, var: AST, new: AST) -> AST:
    """replace all variables that are equal to var with the stamenent in new"""

    def replace(orig: AST) -> AST:
        if orig == var:
            return new
        return orig

    return transform_ast(lit, "Variable", replace)


def inline_replace_stms(lits: list[AST], var: AST, new: AST) -> list[AST]:
    """replace all variables that are equal to var with the stamenent in new"""
    return [inline_replace_stm(lit, var, new) for lit in lits]


def inline_rule(stm: AST) -> AST:
    """inline most X=... equlities in rules/minimize statements"""
    if stm.ast_type not in (ASTType.Rule, ASTType.Minimize):
        return stm
    for blit in stm.body:
        res = _equality(blit)
        if res is not None:
            var, rest = res
            if [x.name for x in collect_ast(stm, "Variable")].count(var.name) > 1:
                break
    else:
        return stm
    new_body = inline_replace_stms([x for x in stm.body if x != blit], var, rest)
    stm = stm.update(body=new_body)
    if stm.ast_type == ASTType.Rule:
        return inline_rule(stm.update(head=inline_replace_stm(stm.head, var, rest)))
    assert stm.ast_type == ASTType.Minimize
    stm = stm.update(weight=inline_replace_stm(stm.weight, var, rest))
    stm = stm.update(priority=inline_replace_stm(stm.priority, var, rest))
    new_terms: list[AST] = []
    for t in stm.terms:
        new_terms.append(inline_replace_stm(t, var, rest))
    stm = stm.update(terms=new_terms)
    return inline_rule(stm)


def inline_aggregate(stm: AST, globals_: set[AST]) -> AST:
    """inline most X=... equlities in body aggregates literals"""
    if not is_body_aggregate(stm):
        return stm
    for elem in stm.atom.elements:
        for c in elem.condition:
            res = _equality(c)
            if res is not None and res[0] not in globals_:
                var, rest = res
                break
        else:
            continue
        new_conds = inline_replace_stms([x for x in elem.condition if x != c], var, rest)
        new_terms = inline_replace_stms(elem.terms, var, rest)
        new_elements = []
        for e in stm.atom.elements:
            if e != elem:
                new_elements.append(e)
            else:
                new_elements.append(e.update(terms=new_terms, condition=new_conds))
        return inline_aggregate(stm.update(atom=stm.atom.update(elements=new_elements)), globals_)
    return stm


def inline_conditional(stm: AST, globals_: set[AST]) -> AST:
    """inline most X=... equlities in conditional literals"""
    if not is_conditional(stm):
        return stm
    for c in stm.condition:
        res = _equality(c)
        if res is not None and res[0] not in globals_:
            var, rest = res
            break
    else:
        return stm
    new_conds = inline_replace_stms([x for x in stm.condition if x != c], var, rest)
    new_lit = inline_replace_stm(stm.literal, var, rest)
    return inline_conditional(stm.update(literal=new_lit, condition=new_conds), globals_)


def inline_aggregates(stm: AST) -> AST:
    """inline most X=... equlities in all body aggregates of this rule/minimize statement"""
    if stm.ast_type not in (ASTType.Rule, ASTType.Minimize):
        return stm
    return stm.update(body=[inline_aggregate(blit, global_vars_inside_body(stm.body)) for blit in stm.body])


def inline_conditionals(stm: AST) -> AST:
    """inline most X=... equlities in all body conditiona literals of this rule/minimize statement"""
    if stm.ast_type not in (ASTType.Rule, ASTType.Minimize):
        return stm
    return stm.update(body=[inline_conditional(blit, global_vars_inside_body(stm.body)) for blit in stm.body])


def inline_arithmetic(prg: Iterable[AST]) -> list[AST]:
    """inline most X=... equlities in the program"""
    new_prg: list[AST] = []
    for stm in prg:
        stm = inline_rule(stm)
        stm = inline_aggregates(stm)
        stm = inline_conditionals(stm)
        new_prg.append(stm)
    return new_prg


def preprocess(prg: Iterable[AST]) -> list[AST]:
    """preprocess program"""
    return normalize(prg)


def postprocess(prg: Iterable[AST]) -> list[AST]:
    """postprocess program"""
    return inline_arithmetic(prg)



""" does math simplification for FO formulas and aggregates"""
import logging
from collections import OrderedDict, defaultdict
from itertools import chain
from math import lcm
from typing import Any, Optional, cast

import clingo
from clingo import SymbolType
from clingo.ast import (
    AST,
    AggregateFunction,
    ASTType,
    BinaryOperation,
    BinaryOperator,
    BodyAggregateElement,
    BooleanConstant,
    Comparison,
    ComparisonOperator,
    Function,
    Guard,
    Literal,
    Sign,
    SymbolicTerm,
    UnaryOperation,
    UnaryOperator,
)
from sympy import (
    Abs,
    Add,
    Dummy,
    Equality,
    Expr,
    GreaterThan,
    Integer,
    LessThan,
    Mod,
    Mul,
    Number,
    Pow,
    S,
    StrictGreaterThan,
    StrictLessThan,
    Symbol,
    Unequality,
    collect,
    default_sort_key,
    floor,
    groebner,
    oo,
    ordered,
    solve,
)
from sympy.core.numbers import NegativeOne, One, Zero

from ngo.dependency import RuleDependency
from ngo.normalize import exline_arithmetic
from ngo.utils.ast import (
    LOC,
    collect_ast,
    collect_binding_information_body,
    collect_binding_information_head,
    compare,
    conditions_of_body_agg,
    global_vars_inside_body,
    negate_agg,
    negate_comparison,
    rhs2lhs_comparison,
)
from ngo.utils.globals import AGG_STR

log = logging.getLogger(__name__)


class SympyApi(Exception):
    """Exception class if simplified formula is to complex or something went wrong using sympy"""


class MathSimplification:
    """class does math simplification on literals"""

    def __init__(self, prg: list[AST]) -> None:
        self._rdp = RuleDependency(prg)

    @staticmethod
    def cost(body: list[AST]) -> tuple[int, int]:
        """compute a cost for the body given number of aggregates and number of literals"""
        numaggs = 0
        comparisons = 0
        for blit in body:
            numaggs += len(collect_ast(blit, "BodyAggregate"))
            if blit.ast_type == ASTType.Literal and blit.atom.ast_type == ASTType.Comparison:
                comparisons += 1
        return (numaggs, comparisons)

    def execute(  # pylint: disable=too-many-branches, too-many-statements
        self, prg: list[AST], optimize: bool = True
    ) -> list[AST]:
        """return a simplified version of the program"""
        ret: list[AST] = []
        prg = exline_arithmetic(prg)
        newprg = list(prg)
        for oldstm, stm in zip(prg, newprg):
            if stm.ast_type not in (ASTType.Rule, ASTType.Minimize):
                ret.append(oldstm)
                continue
            gb = Goebner()
            newbody: list[AST] = []
            agg_conditions: dict[Sign, set[AST]] = defaultdict(set)
            for blit in stm.body:
                expr_list = gb.to_sympy(blit)
                if expr_list is None:
                    newbody.append(blit)
                    continue
                if blit.ast_type == ASTType.Literal:
                    agg_conditions[blit.sign].update(conditions_of_body_agg(blit.atom))
                gb.equalities[blit] = expr_list
            need_bound: set[AST] = set()
            no_bound_needed: set[AST] = set()
            if stm.ast_type == ASTType.Rule:
                need_bound, no_bound_needed = collect_binding_information_head(stm.head, newbody)
            elif stm.ast_type == ASTType.Minimize:
                need_bound.update(collect_ast(stm.weight, "Variable"))
                need_bound.update(collect_ast(stm.priority, "Variable"))
                for t in stm.terms:
                    need_bound.update(collect_ast(t, "Variable"))
            bound_body, unbound_body = collect_binding_information_body(newbody)
            needed = set.union(bound_body, unbound_body, need_bound, no_bound_needed)
            unbound = set.union(need_bound, unbound_body) - bound_body
            # add all variables that became local and were global
            allvars = set()
            for blit in newbody:
                allvars.update(set(collect_ast(blit, "Variable")))
            needed.update((global_vars_inside_body(stm.body) - global_vars_inside_body(newbody)) & allvars)
            try:
                new_conditions = gb.simplify_equalities(needed, unbound)
                for cond in new_conditions:
                    conditions = set(conditions_of_body_agg(cond.atom))
                    if (
                        stm.ast_type != ASTType.Rule
                        or stm.head == Literal(LOC, Sign.NoSign, BooleanConstant(False))
                        or not conditions
                        or conditions.issubset(agg_conditions[Sign.NoSign])
                    ):
                        newbody.append(Literal(LOC, Sign.NoSign, cond.atom))
                    elif conditions.issubset(agg_conditions[Sign.DoubleNegation]):
                        newbody.append(Literal(LOC, Sign.DoubleNegation, cond.atom))
                    elif conditions.issubset(agg_conditions[Sign.Negation]):
                        newbody.append(Literal(LOC, Sign.Negation, negate_agg(cond.atom)))
                    else:
                        raise SympyApi(f"Couldn't preserve dependency graph, skipping {stm}")  # nocoverage

            except SympyApi as err:
                log.info(str(err))
                ret.append(oldstm)
                continue
            except Exception as err:  # pylint: disable=broad-exception-caught
                log.info(
                    f"""Unable to simplfiy because of:\
 {str(err).replace(" contains an element of the set of generators", "")}"""
                )
                ret.append(oldstm)
                continue
            if collect_binding_information_body(newbody)[1]:
                log.info(f"Simplification could not bind all needed variables, skipping {str(oldstm)}")
                ret.append(oldstm)
                continue
            if optimize and self.cost(newbody) >= self.cost(oldstm.body):
                newbody = oldstm.body
            ret.append(stm.update(body=newbody))
        return ret


class Goebner:
    """class does groebner simplification of polynomials"""

    ast2sympy_op = {
        ComparisonOperator.Equal: Equality,
        ComparisonOperator.GreaterEqual: GreaterThan,
        ComparisonOperator.LessEqual: LessThan,
        ComparisonOperator.LessThan: StrictLessThan,
        ComparisonOperator.GreaterThan: StrictGreaterThan,
        ComparisonOperator.NotEqual: Unequality,
    }

    def __init__(self) -> None:
        self._fo_vars: dict[Symbol, AST] = {}
        self.help_neq_vars: dict[Symbol, ComparisonOperator] = OrderedDict()
        self._sym2agg: dict[Symbol, AST] = {}
        self._constants: dict[Symbol, AST] = {}
        self.equalities: dict[AST, list[Expr]] = {}
        # self.to_keep: list[Symbol] = []

    def _to_equality(self, lhs: Expr, op: ComparisonOperator, rhs: Expr) -> Expr:
        if op == ComparisonOperator.Equal:
            return cast(Expr, rhs - lhs)
        aux = Dummy("temp", integer=True)
        self.help_neq_vars[aux] = op
        return cast(Expr, lhs - rhs - aux)

    def _to_sympy_term(self, t: AST) -> Optional[Expr]:
        # pylint: disable=too-many-return-statements
        # pylint: disable=too-many-branches
        if t.ast_type == ASTType.Variable:
            s = Symbol(str(t), integer=True)
            self._fo_vars[s] = t
            return s
        if t.ast_type == ASTType.SymbolicTerm:  # constants
            symbol = t.symbol
            if symbol.type == SymbolType.Number:
                return cast(Expr, Integer(symbol.number))
            if symbol.type == SymbolType.String:
                log.info(f"Can't simplify string operation {t}")
                return None
            if symbol.type == SymbolType.Infimum:
                return cast(Expr, -cast(Expr, oo))
            if symbol.type == SymbolType.Supremum:
                return cast(Expr, oo)
            if symbol.type == SymbolType.Function:
                if not symbol.arguments:
                    s = Symbol(str(t), integer=True)
                    self._constants[s] = t
                    return s
                return None  # nocoverage
            assert False, f"unknown SymbolicTerm {t}"
        if t.ast_type == ASTType.UnaryOperation:
            term = self._to_sympy_term(t.argument)
            if term is None:
                return None
            if t.operator_type == UnaryOperator.Minus:
                return cast(Expr, Number(0) - term)
            if t.operator_type == UnaryOperator.Negation:
                log.info(f"Can't simplify binary operation {t}")
                return None
            if t.operator_type == UnaryOperator.Absolute:
                return cast(Expr, Abs(term))
        if t.ast_type == ASTType.BinaryOperation:
            if t.operator_type in (BinaryOperator.And, BinaryOperator.Or, BinaryOperator.XOr):
                log.info(f"Can't simplify binary operation {t}")
                return None
            lhs = self._to_sympy_term(t.left)
            rhs = self._to_sympy_term(t.right)
            if lhs is None or rhs is None:
                return None
            if t.operator_type == BinaryOperator.Division:
                return cast(Expr, floor(lhs / rhs))
            if t.operator_type == BinaryOperator.Minus:
                return cast(Expr, lhs - rhs)
            if t.operator_type == BinaryOperator.Modulo:
                return cast(Expr, lhs % rhs)
            if t.operator_type == BinaryOperator.Multiplication:
                return cast(Expr, lhs * rhs)
            if t.operator_type == BinaryOperator.Plus:
                return cast(Expr, lhs + rhs)
            if t.operator_type == BinaryOperator.Power:
                return lhs**rhs
        if t.ast_type == ASTType.Function:
            if not t.arguments:
                return cast(Expr, Symbol(str(t), integer=True))  # nocoverage # can also be a symbol
        return None

    def _to_sympy_comparison(self, c: tuple[AST, ComparisonOperator, AST], neg: bool) -> Optional[Expr]:
        lhs = self._to_sympy_term(c[0])
        if lhs is None:
            return None
        rhs = self._to_sympy_term(c[2])
        if rhs is None:
            return None
        op: ComparisonOperator = negate_comparison(c[1]) if neg else c[1]
        return self._to_equality(lhs, op, rhs)

    def _to_sympy_bodyaggregate(self, agg: AST, neg: bool) -> Optional[list[Expr]]:
        assert agg.ast_type == ASTType.BodyAggregate
        assert agg.left_guard is not None
        ret = []
        nonnegative = None
        if agg.ast_type == ASTType.BodyAggregate and agg.function == AggregateFunction.SumPlus:
            nonnegative = True
        dummy = Dummy(f"agg{agg.location.begin.column}", integer=True, nonnegative=nonnegative)
        op: ComparisonOperator = negate_comparison(agg.left_guard.comparison) if neg else agg.left_guard.comparison
        lhs = self._to_sympy_term(agg.left_guard.term)
        if lhs is None:
            return None
        ret.append(self._to_equality(lhs, op, dummy))
        if agg.right_guard is not None:
            op = negate_comparison(agg.right_guard.comparison) if neg else agg.right_guard.comparison
            rhs = self._to_sympy_term(agg.right_guard.term)
            if rhs is None:
                return None
            ret.append(self._to_equality(dummy, op, rhs))
        self._sym2agg[dummy] = agg.update(left_guard=None, right_guard=None)
        return ret

    def to_sympy(self, ast: AST) -> Optional[list[Expr]]:
        """transform a literal into a list of sympy expressions (that are equal 0)
        or None if too complex
        Inequalities are transformed to equalities with an aux variable"""
        if ast.ast_type == ASTType.Literal:
            sign = ast.sign
            atom = ast.atom
            ret: Optional[list[Expr]]
            if atom.ast_type == ASTType.Comparison:
                c = (atom.term, atom.guards[0].comparison, atom.guards[0].term)
                rel = self._to_sympy_comparison(c, sign == Sign.Negation)
                if rel is None:
                    return None
                return [rel]
            if atom.ast_type == ASTType.BodyAggregate:
                ret = self._to_sympy_bodyaggregate(atom, sign == Sign.Negation)
                if ret is None:
                    return None
                return ret
        # Dummy,
        return None

    def new_sum(self, asts: list[AST]) -> AST:
        """given a list of terms and aggregates, create the sum using clingo AST operations"""
        assert len(asts) >= 2
        aggs = [x for x in asts if x.ast_type == ASTType.BodyAggregate]
        if aggs:  # use next and iteration
            rest = [x for x in asts if x.ast_type != ASTType.BodyAggregate]
            collector = aggs[0]
            if collector.function in (AggregateFunction.Min, AggregateFunction.Max):
                raise SympyApi("Cannot express addition with min/max aggregate, skipping.")

            # add a unique identifier to each aggregate so set semantic does not work over multiple aggregates
            def agg_ident(i: int) -> AST:
                return Function(LOC, AGG_STR, [SymbolicTerm(LOC, clingo.Number(i))], False)

            newelements: list[AST] = []
            for elem in collector.elements:
                newelements.append(elem.update(terms=[*elem.terms, agg_ident(0)]))
            for index in range(1, len(aggs)):
                if aggs[index].function in (AggregateFunction.Min, AggregateFunction.Max):
                    raise SympyApi("Cannot express addition with min/max aggregate, skipping.")  # nocoverage
                for e in aggs[index].elements:
                    newelements.append(e.update(terms=[*e.terms, agg_ident(index)]))
            for index, term in enumerate(rest):
                newelements.append(BodyAggregateElement([term, agg_ident(len(aggs) + index)], []))
            return collector.update(elements=newelements, function=AggregateFunction.Sum)

        collector = asts[0]
        index = 1
        for index in range(1, len(asts)):
            collector = BinaryOperation(LOC, BinaryOperator.Plus, collector, asts[index])
        return collector

    def new_abs(self, asts: list[AST]) -> AST:  # nocoverage # sympy does not seem to support this for polynomials
        """given a list of terms create the abs operation using clingo AST operations"""
        if len(asts) != 1:
            raise SympyApi("Missing Sympy specification for more than one absolute argument, skipping.")  # nocoverage
        aggs = [x for x in asts if x.ast_type == ASTType.BodyAggregate]
        if aggs:
            raise SympyApi("Cannot express absolute of aggregates, skipping.")
        return UnaryOperation(LOC, UnaryOperator.Absolute, asts[0])

    def new_pow(self, asts: list[AST]) -> AST:
        """given a list of terms create the power operation using clingo AST operations"""
        if len(asts) != 2:
            raise SympyApi("Missing Sympy specification for more than one power argument, skipping.")  # nocoverage
        aggs = [x for x in asts if x.ast_type == ASTType.BodyAggregate]
        if aggs:
            raise SympyApi("Cannot express exponentiation of aggregates, skipping.")
        return BinaryOperation(LOC, BinaryOperator.Power, asts[0], asts[1])

    def new_mul(self, asts: list[AST]) -> AST:
        """given a list of terms create the multiplication using clingo AST operations"""
        assert len(asts) >= 2
        aggs = [x for x in asts if x.ast_type == ASTType.BodyAggregate]
        if aggs:  # use next and iteration
            if len(aggs) > 1:
                raise SympyApi("Cannot express multiplication of aggregates, skipping.")
            collector = aggs[0]
            if collector.function in (AggregateFunction.Min, AggregateFunction.Max):
                raise SympyApi("Cannot express multiplication with min/max aggregate, skipping.")
            rest = [x for x in asts if x.ast_type != ASTType.BodyAggregate]
            newelements: list[AST] = []
            factor: AST = rest[0]
            for factor_index in range(1, len(rest)):
                factor = BinaryOperation(LOC, BinaryOperator.Multiplication, factor, rest[factor_index])
            for elem in collector.elements:
                newelem = elem
                if not newelem.terms:
                    newelem = newelem.update(
                        terms=[SymbolicTerm(LOC, clingo.Number(1))]
                    )  # nocoverage # not able to produce with parsing
                newterms = list(newelem.terms)
                newterms[0] = BinaryOperation(LOC, BinaryOperator.Multiplication, newterms[0], factor)
                newelements.append(newelem.update(terms=newterms))
            return collector.update(elements=newelements)

        collector = asts[0]
        index = 1
        for index in range(1, len(asts)):
            collector = BinaryOperation(LOC, BinaryOperator.Multiplication, collector, asts[index])
        return collector

    def sympy2ast(self, expr: Expr) -> AST:
        """convert a sympy expr to a clingo AST"""
        if expr.func in (Integer, Zero, NegativeOne, One):
            assert len(expr.args) == 0
            return SymbolicTerm(LOC, clingo.Number(int(expr)))
        if expr.func in (Symbol, Dummy):
            assert isinstance(expr, Symbol)
            if expr in self._fo_vars:
                return self._fo_vars[expr]
            if expr in self._sym2agg:
                return self._sym2agg[expr]
            if expr in self._constants:
                return self._constants[expr]
            assert False, "Solve for t first ?"

        asts: list[AST] = [self.sympy2ast(cast(Expr, sub)) for sub in expr.args]
        if expr.func == Add:
            return self.new_sum(asts)
        if expr.func == Mul:
            return self.new_mul(asts)
        if expr.func == Pow:
            if len(expr.args) == 2 and expr.args[1].is_positive is True:  # type: ignore
                return self.new_pow(asts)
            raise SympyApi("Division by negative pow not supported, skipping.")
        if expr.func == Abs:  # nocoverage # sympy has problams with abs
            return self.new_abs(asts)
        if expr.func == Mod:  # nocoverage # sympy has problams with mod
            raise SympyApi("Modulo not supported, skipping.")
        raise SympyApi(f"Not Implemented conversion {expr.func}")  # e.g. rationals

    @staticmethod
    def is_const(lhs: Expr) -> bool:
        """check if sympy expression is a constant integer"""
        return lhs.is_number is True

    def relation2ast(self, lhs: Expr, op: ComparisonOperator, rhs: Expr) -> AST:
        """lhs is either variable for aggregate, fo_variable or constant
        rhs is an expr"""
        if self.is_const(lhs) and self.is_const(rhs):
            return BooleanConstant(compare(int(lhs), op, int(rhs)))
        rhs_ast = self.sympy2ast(rhs)
        if rhs_ast.ast_type == ASTType.BodyAggregate:
            return rhs_ast.update(left_guard=Guard(op, self.sympy2ast(lhs)))
        lhs_ast = self.sympy2ast(lhs)
        return Comparison(lhs_ast, [Guard(op, rhs_ast)])

    def double_relation2ast(
        self, lhs: Expr, opl: ComparisonOperator, mid: Expr, opr: ComparisonOperator, rhs: Expr
    ) -> AST:
        """lhs is either variable for aggregate, fo_variable or constant
        rhs is an expr"""
        if self.is_const(mid):
            if self.is_const(lhs):
                if compare(int(lhs), opl, int(mid)):
                    return self.relation2ast(mid, opr, rhs)
                return BooleanConstant(False)  # nocoverage, hard to reproduce if not impossible
            if self.is_const(rhs):  # nocoverage
                if compare(int(mid), opr, int(rhs)):
                    return self.relation2ast(lhs, opl, mid)
                return BooleanConstant(False)
        mid_ast = self.sympy2ast(mid)
        if mid_ast.ast_type == ASTType.BodyAggregate:
            return mid_ast.update(
                left_guard=Guard(opl, self.sympy2ast(lhs)), right_guard=Guard(opr, self.sympy2ast(rhs))
            )
        lhs_ast = self.sympy2ast(lhs)  # nocoverage
        rhs_ast = self.sympy2ast(rhs)  # nocoverage
        return Comparison(lhs_ast, [Guard(opl, mid_ast), Guard(opr, rhs_ast)])  # nocoverage

    def remove_unneeded_formulas(self, formulas: list[Expr], needed_symbols: set[Symbol]) -> list[Expr]:
        """filter out formulas that have an unneeded variable on its own"""
        var_stats: dict[Symbol, list[Expr]] = defaultdict(list)  # for each variable, the formulas where it occurs
        for f in formulas:
            for v in cast(set[Symbol], set(f.free_symbols)) & set(self._fo_vars.keys()):
                var_stats[v].append(f)
        ret = list(formulas)
        for v in set(var_stats.keys()) - needed_symbols:
            if len(var_stats[v]) == 1:
                ret.remove(var_stats[v][0])
        return ret

    def least_common(self, rel1: dict[Any, Any], rel2: dict[Any, Any]) -> Optional[tuple[int, int]]:
        """return the factors that both sides need to be multiplied to be equal if possible"""
        agg_keys = [key for key in rel1 if key in rel2 and key.free_symbols.intersection(self._sym2agg.keys())]
        for key in agg_keys:
            if rel1[key].is_constant() or not rel2[key].is_constant():
                least_common = lcm(rel1[key], rel2[key])
                return int(least_common / rel1[key]), int(least_common / rel2[key])
        for key in rel1:
            if key in rel2 and rel1[key].is_constant() and rel2[key].is_constant():
                least_common = lcm(rel1[key], rel2[key])
                return int(least_common / rel1[key]), int(least_common / rel2[key])
        return None

    def combine(
        self, relations: list[tuple[Expr, ComparisonOperator, Expr]], first: int, second: int
    ) -> Optional[tuple[Expr, ComparisonOperator, Expr, ComparisonOperator, Expr]]:
        """combine two relations to a combined one if possible"""
        rel1 = relations[first][2]
        linear1 = collect(rel1, list(self._sym2agg.keys()), evaluate=False, exact=False)
        rel2 = relations[second][2]
        linear2 = collect(rel2, list(self._sym2agg.keys()), evaluate=False, exact=False)
        common = set(linear1.keys()).intersection(linear2.keys())
        agg_common = set()
        for x in sorted(common, key=default_sort_key):
            agg_symbols = set(x.free_symbols).intersection(self._sym2agg.keys())
            if len(agg_symbols) > 0:
                agg_common.update(agg_symbols)

        # if both have a common non empty subset and no other aggregate variables
        if (
            common
            and len((set(rel1.free_symbols) - agg_common).intersection(self._sym2agg.keys())) == 0
            and len((set(rel2.free_symbols) - agg_common).intersection(self._sym2agg.keys())) == 0
        ):
            # both can be a multiple of each other
            factors = self.least_common(linear1, linear2)
            if factors is None:
                return None
            factor1, factor2 = factors

            # keep common factors in the middle, move uncommon ones to the left if possible
            for both in common:
                if linear1[both] * factor1 != linear2[both] * factor2:
                    if both.free_symbols.intersection(self._sym2agg.keys()):
                        return None  # nocoverage
                    relations[first] = (
                        relations[first][0] - (both * linear1[both]),
                        relations[first][1],
                        relations[first][2] - (both * linear1[both]),
                    )
                    relations[second] = (
                        relations[second][0] - (both * linear2[both]),
                        relations[second][1],
                        relations[second][2] - (both * linear2[both]),
                    )

            lhs = relations[first][0] * factor1
            opl = relations[first][1] if factor1 > 0 else rhs2lhs_comparison(relations[first][1])
            mid = relations[first][2] * factor1
            opr = relations[second][1] if factor2 < 0 else rhs2lhs_comparison(relations[second][1])
            rhs = relations[second][0] * factor2
            return (lhs, opl, mid, opr, rhs)
        return None

    def compress_relations(
        self, relations: list[tuple[Expr, ComparisonOperator, Expr]]
    ) -> list[tuple[Expr, ComparisonOperator, Expr] | tuple[Expr, ComparisonOperator, Expr, ComparisonOperator, Expr]]:
        """compress a bunch of ternary relations (lhs, op, rhs) to possibly 5 tuples (lhs, op, mid, opr, rhs)"""
        ret: list[
            tuple[Expr, ComparisonOperator, Expr] | tuple[Expr, ComparisonOperator, Expr, ComparisonOperator, Expr]
        ] = []
        first = 0
        while first < len(relations):
            found: bool = False
            for second, _ in enumerate(relations[first + 1 :], first + 1):
                comb = self.combine(relations, first, second)
                if comb is not None:
                    ret.append(comb)
                    relations.pop(second)
                    relations.pop(first)
                    found = True
                    break

            if not found:
                ret.append(relations[first])
                first += 1

        return ret

    def simplify_equalities(  # pylint: disable=too-many-branches, too-many-statements
        self, needed_vars: set[AST], need_bound: set[AST]
    ) -> list[AST]:
        """Given self.equalities, return a simplified list using the needed variables"""
        assert need_bound.issubset(needed_vars)
        unbound = need_bound - set(self._fo_vars.values())
        if unbound:
            raise SympyApi(
                f"Variables {[str(v) for v in unbound]} seem to be unbound\
 on line {str(next(iter(unbound)).location.begin)}"
            )  # nocoverage
        nothing = list(self.equalities.keys())
        inv_fo = {v: k for k, v in self._fo_vars.items()}
        needed_vars_symbols: set[Symbol] = set(inv_fo[var] for var in needed_vars if var in inv_fo)
        needed_bound_symbols = {inv_fo[var] for var in need_bound}
        varlist = []
        varlist.extend([x for x in self._fo_vars if x not in needed_vars_symbols])
        varlist.extend(self.help_neq_vars.keys())
        varlist.extend(self._sym2agg.keys())
        varlist.extend(ordered(needed_vars_symbols))

        equalities: list[Expr] = [e for xl in self.equalities.values() for e in xl]
        if not varlist or not equalities:
            return nothing
        base = groebner(equalities, varlist)
        ret: list[AST] = []

        simplified_expressions = self.remove_unneeded_formulas(list(base), needed_vars_symbols)

        relations: list[tuple[Expr, ComparisonOperator, Expr]] = []
        # try to bind all needed variables
        solved_for = set()
        for expr in reversed(simplified_expressions):
            free = cast(set[Symbol], set(list(expr.free_symbols)))
            neq_vars = set.intersection(free, self.help_neq_vars.keys())
            if len(neq_vars) > 1:
                return nothing
            if len(neq_vars) == 1:
                v = list(neq_vars)[0]
                lexpr = solve(expr, v)
                if len(lexpr) != 1:
                    return nothing  # nocoverage
                relations.append((S(0), rhs2lhs_comparison(self.help_neq_vars[v]), lexpr[0]))
                continue

            common = list(ordered(set.intersection(free, needed_vars_symbols)))
            if common:
                # solve for some random needed variable
                solved = False
                for solve_for in chain(
                    sorted(needed_bound_symbols, key=default_sort_key),
                    sorted(needed_vars_symbols, key=default_sort_key),
                ):
                    if solve_for in solved_for:
                        continue
                    try:
                        lexpr = solve(expr, solve_for)  # solve for the first one, have to think of a valid strategy
                        if len(lexpr) != 1:  # negative squareroots etc...
                            continue
                        ret.append(
                            Literal(LOC, Sign.NoSign, self.relation2ast(solve_for, ComparisonOperator.Equal, lexpr[0]))
                        )
                        solved_for.add(solve_for)
                        simplified_expressions.remove(expr)
                        solved = True
                        break
                    except Exception:  # pylint: disable=broad-exception-caught
                        continue
                if solved:
                    continue

            # 2. Equality
            if common or not set.intersection(free, self._fo_vars.keys()):
                relations.append((S(0), ComparisonOperator.Equal, expr))

        if not needed_bound_symbols.issubset(solved_for):
            return nothing

        for rel in sorted(self.compress_relations(relations), key=default_sort_key):
            if len(rel) == 3:
                ret.append(Literal(LOC, Sign.NoSign, self.relation2ast(*rel)))
                continue
            ret.append(Literal(LOC, Sign.NoSign, self.double_relation2ast(*rel)))
        return ret



"""
 This module removes unused predicates and arities
"""

import logging
from collections import Counter, defaultdict
from copy import deepcopy
from functools import partial
from itertools import chain
from typing import Callable, Sequence

from clingo.ast import AST, ASTType, Function, Sign, SymbolicAtom, Variable

from ngo.dependency import RuleDependency
from ngo.normalize import exline_arithmetic
from ngo.utils.ast import LOC, Predicate, collect_ast, is_predicate, transform_ast
from ngo.utils.globals import UniqueNames, UniqueVariables

log = logging.getLogger(__name__)


class UnusedTranslator:
    """Removes unecessary predicates and arities from the program"""

    def __init__(self, prg: list[AST], input_predicates: list[Predicate], output_predicates: list[Predicate]):
        self.unique_names = UniqueNames(prg, input_predicates)
        self.input_predicates = input_predicates
        self.output_predicates = output_predicates
        self.used_positions: dict[Predicate, set[int]] = defaultdict(set)
        self.used: set[Predicate] = set()
        self._anon = Variable(LOC, "_")
        self.new_names: dict[tuple[Predicate, Predicate], str] = {}

    @staticmethod
    def transform_body_ast_except_aggregate(stm: AST, ast_type: str, func: Callable[[AST], AST]) -> AST:
        """do call transform on everythin in the body except Aggregate"""
        for index, part in enumerate(stm.body):
            if part.ast_type == ASTType.Literal and part.atom.ast_type in (ASTType.Aggregate, ASTType.BodyAggregate):
                continue
            stm.body[index] = transform_ast(part, ast_type, func)
        return stm

    def _anonymize_variables(self, prg: list[AST]) -> list[AST]:
        new_prg: list[AST] = []

        def anom_var(collection: dict[AST, int], var: AST) -> AST:
            if collection[var] == 1:
                return self._anon
            return var

        for stm in prg:
            if stm.ast_type in (ASTType.Rule, ASTType.Minimize):
                var_collection = Counter(collect_ast(stm, "Variable"))
                stm = self.transform_body_ast_except_aggregate(stm, "Variable", partial(anom_var, var_collection))
            new_prg.append(stm)
        return new_prg

    def _add_usage_stm(self, stm: AST) -> None:
        """add all predicate positions with non anonymous variables to self.usage_xxx"""
        for func in collect_ast(stm, "Function"):
            pred = Predicate(func.name, len(func.arguments))
            self.used.add(pred)
            for index, arg in enumerate(func.arguments):
                if arg != self._anon:
                    self.used_positions[pred].add(index)

    def _add_usage(self, stms: Sequence[AST]) -> None:
        """add all predicate positions with non anonymous variables to self.usage_xxx"""
        for stm in stms:
            self._add_usage_stm(stm)

    def analyze_usage(self, prg: list[AST]) -> None:
        """analyze program which predicates positions are used"""
        for stm in prg:
            if stm.ast_type in (
                ASTType.Rule,
                ASTType.Minimize,
                ASTType.External,
                ASTType.Edge,
                ASTType.Heuristic,
                ASTType.ProjectAtom,
            ):
                self._add_usage(stm.body)
            if stm.ast_type == ASTType.Rule and stm.head.ast_type in (
                ASTType.TheoryAtom,
                ASTType.Disjunction,
                ASTType.HeadAggregate,
                ASTType.Aggregate,
            ):
                for elem in stm.head.elements:
                    self._add_usage(elem.condition)
            if stm.ast_type == ASTType.Rule and stm.head.ast_type in (
                ASTType.Disjunction,
                ASTType.Aggregate,
            ):
                for elem in stm.head.elements:
                    self._add_usage_stm(elem.literal)
            if stm.ast_type in (ASTType.ShowSignature, ASTType.ProjectSignature):
                pred = Predicate(stm.name, stm.arity)
                self.used.add(pred)
                self.used_positions[pred].update(range(0, stm.arity))

        for pred in chain(self.input_predicates, self.output_predicates):
            self.used.add(pred)
            self.used_positions[pred].update(range(0, pred.arity))

    def _new_name(self, orig_pred: Predicate, new_pred: Predicate) -> str:
        key = (orig_pred, new_pred)
        if (orig_pred, new_pred) not in self.new_names:
            self.new_names[key] = self.unique_names.new_predicate(new_pred.name, new_pred.arity).name
            self.used.add(Predicate(self.new_names[key], new_pred.arity))
            log.info(f"Replaced {orig_pred.name}/{orig_pred.arity} with {self.new_names[key]}/{new_pred.arity}.")
        return self.new_names[key]

    def transform(self, atom: AST) -> AST:
        """replace Function inside literal if position is unused"""
        if atom.symbol.ast_type == ASTType.Function:
            term = atom.symbol
            pred = Predicate(term.name, len(term.arguments))
            args = []
            for pos in [x for x in range(0, pred.arity) if x in self.used_positions[pred]]:
                args.append(term.arguments[pos])
            if args != list(term.arguments):
                name = self._new_name(pred, Predicate(term.name, len(args)))
                term = term.update(name=name, arguments=args)
                return atom.update(symbol=term)
        return atom

    def _project_unused_stm(self, stm: AST) -> AST:
        """replace all literal predicates with unused positions with "smaller" predicates in the statement"""
        stm = transform_ast(stm, "SymbolicAtom", self.transform)
        return stm

    def project_unused(self, prg: list[AST]) -> list[AST]:
        """replace all literal predicates with unused positions with "smaller" predicates"""
        new_prg: list[AST] = []
        for stm in prg:
            stm = self._project_unused_stm(stm)
            new_prg.append(stm)
        return new_prg

    def remove_unused(self, prg: list[AST]) -> list[AST]:
        """remove all simple rules that produce unused predicate literals
        Currently skip disjunctions and aggregates in the heads
        """
        new_prg: list[AST] = []
        for stm in prg:
            if (
                stm.ast_type == ASTType.Rule
                and stm.head.ast_type == ASTType.Literal
                and stm.head.sign == Sign.NoSign
                and stm.head.atom.ast_type == ASTType.SymbolicAtom
                and stm.head.atom.symbol.ast_type == ASTType.Function
            ):
                symbol = stm.head.atom.symbol
                pred = Predicate(symbol.name, len(symbol.arguments))
                if pred not in self.used:
                    log.info(f"Remove predicate {pred.name}/{pred.arity} completely.")
                    continue
            new_prg.append(stm)
        return new_prg

    class Mapper:
        """maps a singular rule of the form a(X) :- b(X)."""

        def __init__(self, uv: UniqueVariables, rule_id: int, arguments: list[AST], symbol: AST) -> None:
            self.uv = uv
            self.rule_id = rule_id
            vars_: set[AST] = set()
            for arg in arguments:
                vars_.update(collect_ast(arg, "Variable"))
            map_: dict[AST, AST] = {}
            for v in vars_:
                map_[v] = uv.make_unique(v)

            def replace(var: AST) -> AST:
                if var in map_:
                    return map_[var]
                return var

            self.arguments = [transform_ast(arg, "Variable", replace) for arg in arguments]
            # self.arguments = deepcopy(arguments)
            self.symbol = symbol.update(arguments=[transform_ast(arg, "Variable", replace) for arg in symbol.arguments])
            # self.symbol = deepcopy(symbol)

        def convert(self, arguments: list[AST]) -> AST:
            """places the new arguments inside the symbol (and returns it)
            given by order of self.arguments"""

            def replace(input_: AST, old: AST, new: AST) -> AST:
                if input_ == old:
                    return new
                return input_

            args = deepcopy(list(self.symbol.arguments))
            for index, _ in enumerate(args):
                for head_arg, new_arg in zip(self.arguments, arguments):
                    args[index] = transform_ast(args[index], "Variable", partial(replace, old=head_arg, new=new_arg))

            old_vars: set[AST] = set()
            for arg in arguments:
                old_vars.update(collect_ast(arg, "Variable"))

            def replace_rest(input_: AST, old_vars: set[AST]) -> AST:
                if input_ not in old_vars:
                    return Variable(LOC, "_")
                return input_

            for index, arg in enumerate(args):
                args[index] = transform_ast(arg, "Variable", partial(replace_rest, old_vars=old_vars))
            return SymbolicAtom(Function(LOC, self.symbol.name, args, False))

    def remove_single_copies(self, prg: list[AST]) -> list[AST]:
        """remove rules of the form a(X) :- b(X) and replaces a/1 with b/1"""
        ret: list[AST] = []
        # create a mapping for all 1to1 replaceable predicates
        mapping: dict[Predicate, "UnusedTranslator.Mapper"] = {}
        rd = RuleDependency(prg)

        for head in rd.get_headderivable_predicates():
            if head in self.input_predicates or head in self.output_predicates:
                continue

            rules = rd.get_rules_that_derive(head)
            if not len(rules) == 1:
                continue
            hlit = rules[0].head
            if not (is_predicate(hlit) and hlit.sign == Sign.NoSign):
                continue
            if not len(rules[0].body) == 1:
                continue
            blit = rules[0].body[0]
            if not (is_predicate(blit) and blit.sign == Sign.NoSign):
                continue
            # very simple
            if not all(map(lambda x: x.ast_type == ASTType.Variable, hlit.atom.symbol.arguments)):
                continue
            if not len(hlit.atom.symbol.arguments) == len(blit.atom.symbol.arguments):
                continue
            if head == Predicate(blit.atom.symbol.name, len(blit.atom.symbol.arguments)):
                continue
            mapping[head] = UnusedTranslator.Mapper(
                UniqueVariables(rules[0]), prg.index(rules[0]), list(hlit.atom.symbol.arguments), blit.atom.symbol
            )

        used: set[int] = set()

        def convert(atom: AST) -> AST:
            bpred = Predicate(atom.symbol.name, len(atom.symbol.arguments))
            if bpred in mapping:
                used.add(mapping[bpred].rule_id)
                return mapping[bpred].convert(atom.symbol.arguments)
            return atom

        for rule in prg:
            ret.append(transform_ast(rule, "SymbolicAtom", convert))
        return [x for i, x in enumerate(ret) if i not in used]

    def execute(self, prg: list[AST]) -> list[AST]:
        """
        remove all predicates that are unconstrained and
        additionally do not change the number of answer sets
        """
        new_prg: list[AST] = []
        new_prg = exline_arithmetic(prg)
        prg = new_prg
        prg = self._anonymize_variables(prg)
        self.analyze_usage(prg)
        prg = self.project_unused(prg)
        prg = self.remove_unused(prg)
        prg = self.remove_single_copies(prg)
        return prg



"""
 This module replaces bodys with symmetry breaking rules of the form X1 != X2
 with X1 < X2 if this preserves semantics.
"""

import logging
from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from itertools import chain, combinations, pairwise
from typing import Any, Iterable, Iterator, Optional

import networkx as nx
from clingo.ast import (
    AST,
    AggregateFunction,
    ASTType,
    BodyAggregate,
    BodyAggregateElement,
    Comparison,
    ComparisonOperator,
    Function,
    Guard,
    Literal,
    Rule,
    Sign,
    SymbolicAtom,
    SymbolicTerm,
    Variable,
)
from clingo.symbol import Number

from ngo.dependency import DomainPredicates, RuleDependency
from ngo.utils.ast import (
    LOC,
    Predicate,
    collect_ast,
    global_vars_inside_body,
    global_vars_inside_head,
    is_predicate,
    largest_subset,
    replace_simple_assignments,
)
from ngo.utils.globals import UniqueNames

log = logging.getLogger(__name__)


class SymmetryTranslator:
    """Translates some min/max aggregates into chains"""

    @dataclass(order=True)
    class Symmetry:
        """a single symmetry"""

        literals: tuple[AST, ...]
        strict_neq: dict[int, list[AST]]  # strict_neq[pos] = [lit, ...] for != comparisons
        nstrict_neq: dict[int, list[AST]]  # nstrict_neq[pos] = [lit, ...] for < comparisons

    class SymmetryBundle:
        """a bundle of symmetries that affects each other"""

        def __init__(
            self,
            domain_predicates: DomainPredicates,
            unique_names: UniqueNames,
            in_aggregate: bool,
            symmetries: list["SymmetryTranslator.Symmetry"],
        ):
            self.domain_predicates = domain_predicates
            self.unique_names = unique_names
            self._remove_lits: set[AST] = set()
            self._aux_rules: list[AST] = []
            self._add_lits: set[AST] = set()
            complex_ = len(symmetries) == 1
            for sym in symmetries:
                complex_ = complex_ and bool(
                    len(sym.nstrict_neq) + len(sym.strict_neq) == 1
                )  #  will be false if one sym needs to be simple
            if complex_:
                self.init_complex(in_aggregate, symmetries)
            else:
                self.init_simple(symmetries)

        def empty(self) -> bool:
            """true if nothing is to change"""
            return len(self._add_lits) + len(self._aux_rules) + len(self._remove_lits) == 0

        def init_complex(self, in_aggregate: bool, symmetries: list["SymmetryTranslator.Symmetry"]) -> None:
            """translate to count aggregate"""
            for sym in symmetries:
                # remove symmetries
                for s in sym.literals:
                    self._remove_lits.add(s)
                # remove inequalities
                for lits in chain(sym.strict_neq.values(), sym.nstrict_neq.values()):
                    self._remove_lits.update(lits)
                if in_aggregate:
                    # create new rule and extend condition with new predicate
                    aux_body = self._create_count(sym, self._aux_rules)
                    args = [x for x in aux_body[0].atom.symbol.arguments if x.name != "_"]
                    pred = self.unique_names.new_auxpredicate(len(args))
                    head_lit = Literal(LOC, Sign.NoSign, SymbolicAtom(Function(LOC, pred.name, args, False)))
                    self._aux_rules.append(Rule(LOC, head_lit, aux_body))
                    self._add_lits.add(head_lit)
                else:
                    self._add_lits.update(self._create_count(sym, self._aux_rules))

        def init_simple(self, symmetries: list["SymmetryTranslator.Symmetry"]) -> None:
            """simply improve symmetry by replacing one != with <"""
            improve = True
            for sym in symmetries:
                if sym.nstrict_neq:
                    improve = False

            if improve:
                # Improve symmetries by changing 1 set of variable comparisons
                sym = symmetries[0]
                it = next(iter(sym.strict_neq))
                collect: set[AST] = set()
                for lit in sym.strict_neq[it]:
                    self._remove_lits.add(lit)
                    collect.add(lit.atom.term)
                    collect.add(lit.atom.guards[0].term)
                for lhs, rhs in pairwise(sorted(collect)):
                    self._add_lits.add(
                        Literal(LOC, Sign.NoSign, Comparison(lhs, [Guard(ComparisonOperator.LessThan, rhs)]))
                    )

        def _create_count(self, symmetry: "SymmetryTranslator.Symmetry", rules: list[AST]) -> list[AST]:
            """given a symmetry, create the representive count aggregate and projector and return it
            might add aux rules to the program"""
            # create projected lit
            uneq_positions = set(chain(symmetry.strict_neq.keys(), symmetry.nstrict_neq.keys()))
            same: list[AST] = []
            nsame: list[AST] = []
            first_sym = symmetry.literals[0]
            for i, x in enumerate(first_sym.atom.symbol.arguments):
                if i in uneq_positions:
                    nsame.append(x)
                    same.append(Variable(LOC, "_"))
                else:
                    same.append(x)
            # get domain if possible
            name = first_sym.atom.symbol.name
            pred = Predicate(name, len(same))
            if self.domain_predicates.has_domain(pred):
                rules.extend(self.domain_predicates.create_domain(pred))
                name = self.domain_predicates.domain_predicate(pred).name
            symbol = first_sym.atom.symbol.update(name=name, arguments=same)
            atom = first_sym.atom.update(symbol=symbol)

            # create new aggregate
            elements = [BodyAggregateElement(nsame, [first_sym])]
            agg = BodyAggregate(
                LOC,
                Guard(ComparisonOperator.LessEqual, SymbolicTerm(LOC, Number(len(symmetry.literals)))),
                AggregateFunction.Count,
                elements,
                None,
            )
            return [first_sym.update(atom=atom), Literal(LOC, Sign.NoSign, agg)]

        def remove_lits(self) -> list[AST]:
            """a list of all literals that neede to be removed"""
            return sorted(self._remove_lits)

        def aux_rules(self) -> list[AST]:
            """a list of aux rules that need to be added"""
            return self._aux_rules

        def add_lits(self) -> list[AST]:
            """a list of all literals that need to be added"""
            return sorted(self._add_lits)

    def __init__(self, prg: list[AST], input_predicates: list[Predicate]):
        self.unique_names = UniqueNames(prg, input_predicates)
        self.rule_dependency = RuleDependency(prg)
        self.domain_predicates = DomainPredicates(self.unique_names, prg)

    Inequalities = dict[ComparisonOperator, list[tuple[AST, AST, AST]]]

    @staticmethod
    def _unequal(lhs: AST, rhs: AST, inequalities: Inequalities) -> Optional[tuple[ComparisonOperator, AST]]:
        """given two terms lhs and rhs and some inequalities
        return (OP, lit) if they are unequal or None"""
        for op, ineqs in inequalities.items():
            for lit, var1, var2 in ineqs:
                if (lhs == var1 and rhs == var2) or (lhs == var2 and rhs == var1):
                    return (op, lit)
        return None

    def _crosscheck(
        self,
        potential_equalities: list[set[AST]],
        potential_strict_inequalities: list[dict[int, list[AST]]],
        potential_nstrict_inequalities: list[dict[int, list[AST]]],
        lits_param: list[AST],
        global_vars: set[AST],
        in_aggregate: bool,
    ) -> Iterator[SymmetryBundle]:
        for index_subset in largest_subset(range(0, len(potential_equalities))):
            used_variables: set[AST] = set()
            used_uneq_variables: dict[int, set[AST]] = defaultdict(set)
            lits: list[AST] = list(lits_param)
            for index in index_subset:
                for lit in potential_equalities[index]:
                    lits.remove(lit)
                for pos, neq_lits in chain(
                    potential_strict_inequalities[index].items(), potential_nstrict_inequalities[index].items()
                ):
                    lits = [x for x in lits if x not in neq_lits]
                    for pred in potential_equalities[index]:
                        used_variables.update(collect_ast(pred.atom.symbol.arguments[pos], "Variable"))
                        used_uneq_variables[index].update(collect_ast(pred.atom.symbol.arguments[pos], "Variable"))
            if len((global_vars_inside_body(lits) | global_vars) & used_variables) == 0:
                # built ccs, in a cc, only one comparison can be improved
                g = nx.Graph()
                for index1 in index_subset:
                    g.add_edge(index1, index1)  # cc's of length 1
                    for index2 in index_subset:
                        if index1 == index2:
                            continue
                        if used_uneq_variables[index1] & used_uneq_variables[index2]:
                            g.add_edge(index1, index2)
                for cc in nx.connected_components(g):
                    yield SymmetryTranslator.SymmetryBundle(
                        self.domain_predicates,
                        self.unique_names,
                        in_aggregate,
                        [
                            SymmetryTranslator.Symmetry(
                                literals=tuple(sorted(potential_equalities[index])),
                                strict_neq=potential_strict_inequalities[index],
                                nstrict_neq=potential_nstrict_inequalities[index],
                            )
                            for index in cc
                        ],
                    )
                return

    def largest_symmetric_group(
        self, body: list[AST], global_vars: set[AST], rest: list[AST], in_aggregate: bool
    ) -> Iterator[SymmetryBundle]:
        """
        Given a list A of literals (a body or a condition)
        Given a list B of literals that is also visible in the same scope
        return tuples of literals with the same predicate from list A
        and a list of inequalities used (from A)
        All Variables used in the same predicates are either equal
        or unequal (with proof in the inequalities)
        Only one asymetric inequality is allowed.
        All unequal Variables are not to allowed to be used within the rest of the
        visible level of A and B
        At least one inequality needs to be present.
        """
        potential_equalities: list[set[AST]] = []
        potential_strict_inequalities: list[dict[int, list[AST]]] = []
        potential_nstrict_inequalities: list[dict[int, list[AST]]] = []
        inequalities = SymmetryTranslator._inequalities(body)

        # for each equality
        #     check if it is not overlapping an already existing potential equality
        #     do all tests and collect a set of potential equalities
        #     Then do a crosscheck of the uneqal variables of all of them
        #     None of them may occur outside
        #     If they occur outside, only take a subset of the potential equalities
        # Group them together by shared variables
        def intersects(sub: set[Any], dom: set[Any]) -> bool:
            return len(sub.intersection(dom)) > 0

        for equality in SymmetryTranslator._all_equal_symbols(body):
            if any(map(partial(intersects, set(equality)), potential_equalities)):
                continue
            used_inequalities = []
            fine = True
            for pos, sides in enumerate(zip(*[x.atom.symbol.arguments for x in equality])):
                # (lhs, rhs) = sides
                # 1. all sides are equal
                if len(set(sides)) == 1:
                    continue
                # 2. all sides are inequal
                for lhs, rhs in combinations(sides, 2):
                    res = SymmetryTranslator._unequal(lhs, rhs, inequalities)
                    if res is None:
                        fine = False
                        break
                    op, lit = res
                    used_inequalities.append((op, lit, pos))
                if not fine:
                    break
            if fine and len(used_inequalities) > 0:
                potential_equalities.append(set(equality))
                potential_strict_inequalities.append(defaultdict(list))
                potential_nstrict_inequalities.append(defaultdict(list))
                for op, lit, pos in used_inequalities:
                    if op == ComparisonOperator.NotEqual:
                        potential_strict_inequalities[-1][pos].append(lit)
                    else:
                        potential_nstrict_inequalities[-1][pos].append(lit)
        yield from self._crosscheck(
            potential_equalities,
            potential_strict_inequalities,
            potential_nstrict_inequalities,
            body + rest,
            global_vars,
            in_aggregate,
        )

    @staticmethod
    def _inequalities(body: Iterable[AST]) -> Inequalities:
        """given some literals
        returns inequlities of the form [OP] =[(orignal_literal, variable1, variable2),...]
        expects only simple comparisons
        OP in {!=, <}"""
        ret: SymmetryTranslator.Inequalities = defaultdict(list)
        for lit in body:
            if lit.ast_type == ASTType.Literal and lit.atom.ast_type == ASTType.Comparison:
                assert len(lit.atom.guards) == 1
                atom = lit.atom
                guard = atom.guards[0]
                if atom.term.ast_type != ASTType.Variable or guard.term.ast_type != ASTType.Variable:
                    continue
                if (lit.sign == Sign.NoSign and guard.comparison == ComparisonOperator.NotEqual) or (
                    lit.sign == Sign.Negation and guard.comparison == ComparisonOperator.Equal
                ):
                    ret[ComparisonOperator.NotEqual].append((lit, atom.term, guard.term))
                elif (lit.sign == Sign.NoSign and guard.comparison == ComparisonOperator.LessThan) or (
                    lit.sign == Sign.Negation and guard.comparison == ComparisonOperator.GreaterThan
                ):
                    ret[ComparisonOperator.LessThan].append((lit, atom.term, guard.term))
                elif (lit.sign == Sign.NoSign and guard.comparison == ComparisonOperator.GreaterThan) or (
                    lit.sign == Sign.Negation and guard.comparison == ComparisonOperator.LessThan
                ):
                    ret[ComparisonOperator.LessThan].append((lit, guard.term, atom.term))
        return ret

    @staticmethod
    def _all_equal_symbols(body: Iterable[AST]) -> Iterator[tuple[AST, ...]]:
        """return all tuples of any length of literals that use the same predicate
        can contain subsets, ordered by size"""
        symbols: list[AST] = []
        for lit in body:
            if is_predicate(lit):
                symbols.append(lit)
        for subset in largest_subset(symbols):
            if len(subset) <= 1:
                continue
            preds: set[Predicate] = set()
            for lit in subset:
                symbol = lit.atom.symbol
                preds.add(Predicate(symbol.name, len(symbol.arguments)))
            if len(preds) == 1:
                yield tuple(sorted(subset))

    def _process_aggregates(self, stm: AST) -> list[AST]:
        """given a stm, process all aggregates for symmetries
        and either break them or create aux rules using counting
        """
        # pylint: disable=too-many-nested-blocks
        ret: list[AST] = []
        newbody: list[AST] = []
        for blit in stm.body:
            if not (blit.ast_type == ASTType.Literal and blit.atom.ast_type == ASTType.BodyAggregate):
                newbody.append(blit)
                continue
            new_elements: list[AST] = []
            for elem in blit.atom.elements:
                condition = list(elem.condition)
                global_vars: set[AST] = set()
                if stm.ast_type == ASTType.Rule:
                    global_vars = global_vars_inside_head(stm.head)
                else:
                    for t in [stm.weight, stm.priority, *stm.terms]:
                        global_vars.update(collect_ast(t, "Variable"))
                for symmetry_bundle in list(
                    self.largest_symmetric_group(condition, global_vars, list(elem.terms) + list(stm.body), True)
                ):
                    log.info(f"Replace atleast2 in aggregate {str(blit)}.")
                    for lit in symmetry_bundle.remove_lits():
                        condition.remove(lit)
                    for lit in symmetry_bundle.add_lits():
                        condition.append(lit)
                    ret.extend(symmetry_bundle.aux_rules())

                new_elements.append(elem.update(condition=condition))
            atom = blit.atom.update(elements=new_elements)
            newbody.append(blit.update(atom=atom))

        ret.append(stm.update(body=newbody))
        return ret

    def _process_stm(self, stm: AST) -> list[AST]:
        """This modtrait replaces bodys with symmetry breaking rules that
        count different occurences of the same predicate.
        This can reduce the number of grounded rules.
        Might also return additional aux rules."""
        ret: list[AST] = []
        global_vars: set[AST] = set()
        if stm.ast_type == ASTType.Rule:
            global_vars = global_vars_inside_head(stm.head)
        else:
            for t in [stm.weight, stm.priority, *stm.terms]:
                global_vars.update(collect_ast(t, "Variable"))
        body: list[AST] = list(stm.body)

        for symmetry_bundle in list(self.largest_symmetric_group(body, global_vars, [], False)):
            ### Translate to aggregate
            if not symmetry_bundle.empty():
                log.info(f"Replace atleast2 in {str(stm)}")
                for lit in symmetry_bundle.remove_lits():
                    body.remove(lit)
                for lit in symmetry_bundle.add_lits():
                    body.append(lit)
                ret.extend(symmetry_bundle.aux_rules())

        ret.append(stm.update(body=body))
        return ret

    def _process(self, stm: AST) -> list[AST]:
        """This modtrait replaces bodys with symmetry breaking rules that
        count different occurences of the same predicate with a counting aggregate
        if this preserves semantics. This can reduce the number of grounded
        rules.
        Might also return additional aux rules."""
        ret: list[AST] = []
        stms = self._process_aggregates(stm)
        for r in stms:
            ret.extend(self._process_stm(r))
        return ret

    def execute(self, orig: list[AST]) -> list[AST]:
        """
        replace easy min/max aggregates with chaining rules
        also replace the usage of the results in sum and optimize conditions
        """
        ret: list[AST] = []
        prg: list[AST] = []
        for rule in orig:
            if rule.ast_type in (ASTType.Rule, ASTType.Minimize):
                prg.append(replace_simple_assignments(rule))
                continue
            prg.append(rule)
        for rule in prg:
            if rule.ast_type in (ASTType.Rule, ASTType.Minimize):
                ret.extend(self._process(rule))
                continue
            ret.append(rule)
        return ret



"""
 This module replaces sets of literals that occur in multiple rules
 with an extra predicate that is derived only once
"""

from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations
from typing import Iterable, Optional

import networkx as nx
from clingo.ast import AST, ASTType, Function, Literal, Rule, Sign, SymbolicAtom

from ngo.dependency import DomainPredicates
from ngo.utils.ast import (
    LOC,
    Predicate,
    collect_binding_information_body,
    global_vars_inside_body,
    replace_assignments,
    transform_ast,
)
from ngo.utils.globals import UniqueNames

log = logging.getLogger(__name__)

AUX_VAR = "__AUX_"


@dataclass
class RuleRebuilder:
    """simple datastorage to recreate a rule after removing parts of it"""

    ruleid: int
    sub_ast: Optional[AST]
    sub_sub_ast: Optional[AST]
    original_literals: tuple[AST, ...]
    new_literals: tuple[AST, ...]
    oldvars2newvars: dict[str, str]
    newvars2oldvars: dict[str, str]


class LiteralCollector:
    """read in prg and find duplicates of a certain size,
    changes the program and produces additional rules with an index into the original program where to put them"""

    def __init__(self, size: int, prg: list[AST], additional_rules: dict[int, list[AST]]):
        self.size = size
        self.prg = prg
        self.additional_rules = additional_rules
        self.occurences: dict[tuple[AST, ...], list[RuleRebuilder]] = defaultdict(list)
        for index, stm in enumerate(self.prg):
            if stm.ast_type == ASTType.Rule:
                self._add_occurences_from_body(stm.body, index)
                self._add_occurences_from_conditionals(stm.body, index)
                self._add_occurences_from_body_aggregate(stm.body, index)
            elif stm.ast_type == ASTType.Minimize:
                self._add_occurences_from_body(stm.body, index)
        self._filter_occurences()

    def _filter_occurences(self) -> None:
        remove: list[tuple[AST, ...]] = []
        # connect the global variables of each element with each other.
        # in the end there should be only one scc
        for subset in self.occurences.keys():
            graph = nx.Graph()
            all_vars: set[AST] = set()
            for lit in subset:
                vars_ = global_vars_inside_body([lit])
                all_vars.update(vars_)
                graph.add_edges_from(combinations(vars_, 2))
            cc = list(nx.connected_components(graph))
            if (len(cc) > 1) or (len(cc) == 0 and len(all_vars) > 1):
                remove.append(subset)
                continue
            if len(cc) == 1 and set(cc[0]) < all_vars:
                remove.append(subset)
        for subset in remove:
            del self.occurences[subset]

    def _add_occurences_from_body(self, body: Iterable[AST], index: int) -> None:
        """add all combinations of self.size from literals from the body"""
        for original_subset in combinations(body, self.size):
            _, unbound = collect_binding_information_body(original_subset)
            if not unbound:
                new_subset, oldvars2newvars = anonymize_variables(original_subset)
                newvars2oldvars = {v: k for k, v in oldvars2newvars.items()}
                self.occurences[tuple(new_subset)].append(
                    RuleRebuilder(
                        index, None, None, original_subset, tuple(new_subset), oldvars2newvars, newvars2oldvars
                    )
                )

    def _add_occurences_from_conditionals(self, body: Iterable[AST], index: int) -> None:
        """add all combinations of self.size from literals from a conditional"""
        for lit in body:
            if lit.ast_type == ASTType.ConditionalLiteral:
                for original_subset in combinations(lit.condition, self.size):
                    _, unbound = collect_binding_information_body(original_subset)
                    if not unbound:
                        new_subset, oldvars2newvars = anonymize_variables(original_subset)
                        newvars2oldvars = {v: k for k, v in oldvars2newvars.items()}
                        self.occurences[tuple(new_subset)].append(
                            RuleRebuilder(
                                index, lit, None, original_subset, tuple(new_subset), oldvars2newvars, newvars2oldvars
                            )
                        )

    def _add_occurences_from_body_aggregate(self, body: Iterable[AST], index: int) -> None:
        """add all combinations of self.size from literals from a conditional inside a bodyaggregate"""
        for lit in body:
            if lit.ast_type == ASTType.Literal and lit.atom.ast_type == ASTType.BodyAggregate:
                for element in lit.atom.elements:
                    assert element.ast_type == ASTType.BodyAggregateElement
                    for original_subset in combinations(element.condition, self.size):
                        _, unbound = collect_binding_information_body(original_subset)
                        if not unbound:
                            new_subset, oldvars2newvars = anonymize_variables(original_subset)
                            newvars2oldvars = {v: k for k, v in oldvars2newvars.items()}
                            self.occurences[tuple(new_subset)].append(
                                RuleRebuilder(
                                    index,
                                    lit,
                                    element,
                                    original_subset,
                                    tuple(new_subset),
                                    oldvars2newvars,
                                    newvars2oldvars,
                                )
                            )

    def rebuild(self, rule_builder: RuleRebuilder, predicate_name: str, variables: list[AST]) -> list[AST]:
        """takes a RuleRebuilder and returns the new body for the rule
        returns None if rebuilding failed"""
        rule = self.prg[rule_builder.ruleid]
        if not rule_builder.sub_ast:
            new_body = [lit for lit in rule.body if lit not in rule_builder.original_literals]
            new_body.append(Literal(LOC, Sign.NoSign, SymbolicAtom(Function(LOC, predicate_name, variables, False))))

        elif rule_builder.sub_ast.ast_type == ASTType.ConditionalLiteral:
            new_body = [lit for lit in rule.body if lit != rule_builder.sub_ast]
            new_condition = [lit for lit in rule_builder.sub_ast.condition if lit not in rule_builder.original_literals]
            new_condition.append(
                Literal(LOC, Sign.NoSign, SymbolicAtom(Function(LOC, predicate_name, variables, False)))
            )
            new_body.append(rule_builder.sub_ast.update(condition=new_condition))
        elif (
            rule_builder.sub_ast.ast_type == ASTType.Literal
            and rule_builder.sub_ast.atom.ast_type == ASTType.BodyAggregate
        ):
            assert rule_builder.sub_sub_ast is not None
            new_body = [lit for lit in rule.body if lit != rule_builder.sub_ast]
            new_condition = [
                lit for lit in rule_builder.sub_sub_ast.condition if lit not in rule_builder.original_literals
            ]
            new_condition.append(
                Literal(LOC, Sign.NoSign, SymbolicAtom(Function(LOC, predicate_name, variables, False)))
            )
            new_conditions = [clit for clit in rule_builder.sub_ast.atom.elements if clit != rule_builder.sub_sub_ast]
            new_conditions.append(rule_builder.sub_sub_ast.update(condition=new_condition))
            new_aggregate = rule_builder.sub_ast.atom.update(elements=new_conditions)
            new_body.append(rule_builder.sub_ast.update(atom=new_aggregate))
        else:
            assert f"NOT IMPLEMENTED: can not rebuild {rule_builder}"
        return new_body

    def process(self, unique_names: UniqueNames) -> set[int]:
        """actually apply changes to the self.prg and self.additional_rules
        returns the set of rules that has been modified"""
        changed_rules: set[int] = set()
        for literal_set, rulebuilding in self.occurences.items():
            if len(rulebuilding) > 1:
                if any(map(lambda rb: rb.ruleid in changed_rules, rulebuilding)):
                    continue
                # check for overlaps and ignore if nothing is left
                if len({(b.ruleid, b.sub_ast, b.sub_sub_ast) for b in rulebuilding}) <= 1:
                    continue
                # create new aux predicate
                min_index = min(map(lambda rb: rb.ruleid, rulebuilding))
                bound: list[AST] = sorted(collect_binding_information_body(literal_set)[0])
                aux_pred = unique_names.new_auxpredicate(len(bound))
                new_rule = Rule(
                    location=LOC,
                    head=Literal(LOC, Sign.NoSign, SymbolicAtom(Function(LOC, aux_pred.name, bound, False))),
                    body=literal_set,
                )
                self.additional_rules[min_index].append(new_rule)
                # change old rules to use the new predicate
                for rule_builder in rulebuilding:
                    if rule_builder.ruleid in changed_rules:
                        continue
                    changed_rules.add(rule_builder.ruleid)
                    rule = self.prg[rule_builder.ruleid]
                    reverted_bound = unanonymize_variables(bound, rule_builder.newvars2oldvars)
                    new_body = self.rebuild(rule_builder, aux_pred.name, reverted_bound)
                    if new_body:
                        new_rule = rule.update(body=new_body)  # should work for rules and minimize statements
                        self.prg[rule_builder.ruleid] = new_rule
        return changed_rules


def unanonymize_variables(variables: Iterable[AST], mapping: dict[str, str]) -> list[AST]:
    """change variable names back if they are in the mapping
    remove all other variables"""
    for var in variables:
        assert var.ast_type == ASTType.Variable
    return [var.update(name=mapping[var.name]) for var in variables if var.name in mapping]


def anonymize_variables(literals: Iterable[AST]) -> tuple[list[AST], dict[str, str]]:
    """change variable names in literals to generic ones,
    additionally return a mapping from the old variables to the new ones"""
    counter = 0
    old2new: dict[str, str] = {}
    ret: list[AST] = []

    def replace(var: AST) -> AST:
        nonlocal counter
        nonlocal old2new
        assert var.ast_type == ASTType.Variable
        new = AUX_VAR + str(counter)
        if var.name in old2new:
            new = old2new[var.name]
        elif var.name != "_":
            old2new[var.name] = new
            counter += 1
        else:
            new = var.name
        return var.update(name=new)

    for lit in literals:
        ret.append(transform_ast(lit, "Variable", replace))
    return sorted(ret), old2new


class LiteralDuplicationTranslator:
    """Removes duplicated literal sets in the body"""

    def __init__(self, prg: list[AST], input_predicates: list[Predicate]):
        self.unique_names = UniqueNames(prg, input_predicates)
        self.domain_predicates = DomainPredicates(self.unique_names, prg)

    @staticmethod
    def compute_size_from_body(rule: AST) -> int:
        """compute size of rule body"""
        assert rule.ast_type == ASTType.Rule
        return len(rule.body)

    @staticmethod
    def compute_size_from_minimize(stm: AST) -> int:
        """compute size of rule body"""
        assert stm.ast_type == ASTType.Minimize
        return len(stm.body)

    @staticmethod
    def compute_max_size_from_conditionals(rule: AST) -> int:
        """compute maximum size of a conditional in the body"""
        assert rule.ast_type == ASTType.Rule
        max_size = 0
        for lit in rule.body:
            if lit.ast_type == ASTType.ConditionalLiteral:
                max_size = max(max_size, len(lit.condition))
        return max_size

    @staticmethod
    def compute_max_size_from_body_aggregate(rule: AST) -> int:
        """compute maximum size of conditions in body aggregates"""
        assert rule.ast_type == ASTType.Rule
        max_size = 0
        for lit in rule.body:
            if lit.ast_type == ASTType.Literal and lit.atom.ast_type == ASTType.BodyAggregate:
                for element in lit.atom.elements:
                    assert element.ast_type == ASTType.BodyAggregateElement
                    max_size = max(max_size, len(element.condition))
        return max_size

    def execute(self, prg: list[AST]) -> list[AST]:
        """
        collect all possible sets of literals to find common subsets and replace them
        """
        prg = prg.copy()

        # 1. collect all literals in a rule, create all possible subsets of them for this rule
        # 1.1 all variables in the subset must be bounded
        # 1.2 subsets include
        #  - the literals itself
        #  - variable binding independent of the name (equal variables,
        #  variable comparisons, arithmetics with variables)
        #  - a set of variables that in the subset AND used outside the subset
        # 1.3 also collect all subsets of conditionals
        # 2. find a biggest common subset
        # 3. replace this subset with an additional rule and a new predicate
        # 4. restart from 1 until fixpoint
        maxsize = 0
        newprogram: list[AST] = []
        for stm in prg:
            newprogram.append(replace_assignments(stm))
            if stm.ast_type == ASTType.Rule:
                maxsize = max(maxsize, self.compute_size_from_body(newprogram[-1]))
                maxsize = max(maxsize, self.compute_max_size_from_conditionals(newprogram[-1]))
                maxsize = max(maxsize, self.compute_max_size_from_body_aggregate(newprogram[-1]))
            else:
                if stm.ast_type == ASTType.Minimize:
                    maxsize = max(maxsize, self.compute_size_from_minimize(newprogram[-1]))

        size = maxsize
        restore: list[bool] = [True] * len(prg)
        while size > 1:
            additional_rules: dict[int, list[AST]] = defaultdict(list)  # maps to an index to insert the new rules
            lc = LiteralCollector(size, newprogram, additional_rules)
            changed_rules = lc.process(self.unique_names)
            if len(changed_rules) == 0:
                size -= 1
            for index in changed_rules:
                restore[index] = False
            for index in sorted(additional_rules.keys(), reverse=True):
                newprogram[index:index] = additional_rules[index]
                prg[index:index] = additional_rules[index]
                restore[index:index] = [False] * len(additional_rules[index])
        for index, old in enumerate(restore):
            if old:
                newprogram[index] = prg[index]

        return newprogram
